
#include "unique_number.h" 

#include "analysis/dependencies.h"
#include "analysis/dependency_system.h"

#include "firstorder/nnf.h"
#include "firstorder/simplifications.h"
#include "firstorder/replacement.h"
#include "firstorder/antiskolemization.h"
#include "firstorder/factor.h"
#include "firstorder/geometric.h"

#include "statistics/distribution.h"
#include "statistics/timer.h"
#include "statistics/reporter.h"

#include "firstorder/testing.h"

#include "parser/token.h"
#include "parser/parser.h"
#include "parser/errors.h"
#include "parser/tptp.h"

#include "varmatching.h"

#include "varset.h"

#include "rulesystem.h"
#include "model.h"

#include "normalizer.h"

#include "varsubst.h"
#include "element.h" 
#include "predicate.h"
#include "groundatom.h" 
#include "varatom.h"
#include "partial.h"

#include "predicate.stat"
#include "firstorder/function.stat"
#include "prolog/funcpred.stat"

#include <cstdlib>
#include <iostream>

#include "geometric/parser/token.h"
#include "geometric/parser/tokenizer.h"
#include "geometric/parser/parser.h"

#include "geometric/constraints/gcsp.h"
#include "geometric/constraints/subststack.h"
#include "geometric/constraints/refinement.h" 
#include "geometric/constraints/solve.h"
#include "geometric/constraints/resolution.h"


struct parameters
{
   bool tptp_input;
   bool tptp_output;
      // tptp_output is not yet supported.

   bool nonempty;
      // Insist on non-empty models. With nonempty = false,
      // the formula forall x. p(x) -> exists x. p(x) is not provable.

   std::string inputfilename;
      // File from which input is read. If inputfilename = empty, input
      // will be read from standard input. 
  
   std::string includepath;
      // Path from which include files are read.
 
   std::ifstream* inputstream;


   explicit parameters( const std::list< std::string > & pars )
      : tptp_input{ false },
        tptp_output{ false },
	nonempty{ false },
	inputfilename{ "" },
        includepath{ "" },
	inputstream{ nullptr }
   {

      if( pars. size( ) == 0 )
      {
         std::cout << "ERROR: There should be at least one parameter\n\n";
	 exit(0);
      }

      std::list< std::string > :: const_iterator p = pars. begin( ); 
      ++ p;

      while( p != pars. end( ))
      {
         if( *p == "-tptp_input" )
         {
            tptp_input = true;
	    ++ p; 
            goto recognized;
         }

	 if( *p == "-tptp_output" )
         {
            tptp_output = true;
	    ++ p;
	    goto recognized;
         }

	 if( *p == "-nonempty" )
         {
            nonempty = true;
	    ++ p;
	    goto recognized;
         }

	 if( *p == "-inputfile" ) 
         {
            ++ p;
	    if( p == pars. end( ))
            {
               std::cout << "ERROR: expecting filename after -inputfile\n\n";
	       exit(0);
            }

	    if( inputfilename. size( ))
            {
               std::cout << "ERROR: multiple filenames\n\n";
	       exit(0);
            }

	    inputfilename = *(p++);
	    inputstream = new std::ifstream( inputfilename. c_str( ));
            if( ! ( *inputstream ))
            {
               std::cout << "ERROR: ";
	       std::cout << "could not open inputfile ";
	       std::cout << inputfilename<< "\n\n";
	       exit(0);
            }

            goto recognized;
         }

         if( *p == "-include" )
         {
            ++ p;
            if( p == pars. end( ))
            {
               throw std::runtime_error( "expected a path after -include" );
            }

            if( includepath. size( ))
               throw std::runtime_error( "multiple include paths given" );

            includepath = *(p++);
            if( includepath. back( ) != '/' )
               includepath. push_back( '/' );

            goto recognized;
         }

 
	 std::cout << "ERROR: unrecognized parameter: " << *(p++);
	 std::cout << "\n";
	 exit(0);

      recognized:
	 ;
      }
   }


   ~parameters( )
   {
      delete inputstream;
   }


};


std::ostream& operator << ( std::ostream& stream , const parameters& pars )
{
   stream << "Parameters:\n\n";

   stream << "tptp_input =     " << pars. tptp_input << "\n";
   stream << "tptp_output =    " << pars. tptp_output << "\n";
   stream << "nonempty =       " << pars. nonempty << "\n";
   stream << "inputfile =      " << ( pars. inputfilename. size( ) ?
	                               pars. inputfilename :
				       "(stdin)" ) << "\n";
   stream << "includepath =    " << pars. includepath << "\n"; 
   stream << "\n";
   return stream;
}


#include "geometric/signpred.h"
#include "geometric/truthval.h" 
#include "geometric/variable.h"
#include "geometric/element.h"
#include "geometric/atom.h"
#include "geometric/formula.h"
#include "geometric/rule.h"
#include "geometric/inference.h" 
#include "geometric/range.h"
#include "geometric/unification.h"
#include "geometric/substqueue.h"
#include "geometric/set.h"
#include "geometric/permutation.h"
#include "geometric/permutationgenerator.h"
#include "geometric/subst.h"
#include "geometric/matched.h"

#include "geometric/subststack.h"
#include "geometric/false_instance.h"
#include "geometric/false_instance_set.h"
#include "geometric/deduction.h"
#include "geometric/searchstate.h"
#include "geometric/findfalse.h"
#include "geometric/index/simple.h"


geometric::variable makenew( variable v )
{
   geometric::variable res;
   while( v != variable( ))   
   {
      ++ res;
      -- v;
   }
   return res;
}


geometric::varatom makeatom( geometric::signature& sig,
                             const predicate& p, 
                             geometric::truthval t )
{
   if( p == predicate::diseq && t == geometric::truthval::F )
      return geometric::varatom( sig. equality, geometric::truthval::T );

   if( p == predicate::create )
      return geometric::varatom( sig. create, t );
                                                     
   return geometric::varatom( 
              sig. lookup( { p.getstring( ) }, p.getarity( )), t );
}


geometric::varatom makenew( geometric::signature& sig,
                            const varatom& at, geometric::truthval t )
{
   auto res = makeatom( sig, at.p, t );

   unsigned int i = 0;
   for( auto arg = at. a1; arg != at. a2; ++ arg )
   {
      res [ i ++ ] = makenew( *arg ); 
   }
   return res;  
}


int main( int nrargs, char* args [] )
{
   geometric::variable v0;
   auto v1 = v0. next( ); 
   auto v2 = v1. next( ); 
   auto v3 = v2. next( ); 
   auto v4 = v3. next( );
   auto v5 = v4. next( ); 
   auto v6 = v5. next( );
   auto v7 = v6. next( ); 
   auto v8 = v7. next( );
   auto v9 = v8. next( ); 
   auto v10 = v9. next( ); 

   geometric::element e0;
   auto e1 = e0. next( );
   auto e2 = e1. next( ); 
   auto e3 = e2. next( ); 
   auto e4 = e3. next( ); 
   auto e5 = e4. next( ); 
   auto e6 = e5. next( ); 
   auto e7 = e6. next( ); 
   auto e8 = e7. next( ); 
   auto e9 = e8. next( ); 
   auto e10 = e9. next( );

   geometric::signature sig;
   auto P = sig. lookup( { "P" }, 1 );
   auto Q = sig. lookup( { "Q" }, 1 );
   auto R = sig. lookup( { "R" }, 1 );

   auto A = sig. lookup( { "A" }, 1 ); 
   auto B = sig. lookup( { "B" }, 1 );
   auto C = sig. lookup( { "C" }, 1 );
   auto D = sig. lookup( { "D" }, 2 );

   auto p2 = sig. lookup( { "p2" }, 2 );
   auto q2 = sig. lookup( { "q2" }, 2 );
   auto r2 = sig. lookup( { "r2" }, 2 );

   auto p3 = sig. lookup( { "p3" }, 3 );
   auto q3 = sig. lookup( { "q3" }, 3 ); 
   auto r3 = sig. lookup( { "r3" }, 3 );

   auto p4 = sig. lookup( { "p4" }, 4 );
   auto q4 = sig. lookup( { "q4" }, 4 );

   auto S = sig. lookup( { "S" }, 2 );

   // Just run the prover:


   std::cout << "hallo\n\n";

   std::list< std::string > pppp;
   for( unsigned int i = 0; i < nrargs; ++ i )
   {
      pppp. push_back( args [i] );
   }

   parameters pars = parameters( pppp );

   std::cout << "\n";
   std::cout << pars << "\n";

   parser::inputstream is{ 
      pars. inputfilename. size( ) ? 
         * ( pars. inputstream ) : std::cin,
      pars. inputfilename. size( ) ? 
         pars. inputfilename : "STDIN" };

   model m;
   rulesystem rs;
   firstorder::sequent fol;

   bool hasconjecture = false;
   if( pars. tptp_input )
   {
      try
      {
         parser::tptp::readproblem( is, pars. includepath, fol, hasconjecture );
      }
      catch( parser::readerror e )
      {
         std::cout << "ERROR: ";
         std::cout << e << "\n";
         return 0;
      }
      catch( parser::syntaxerror s )
      {
         std::cout << "ERROR: ";
	 std::cout << s << "\n";
	 return 0;
      }
      catch( parser::conversionerror c )
      {
         std::cout << "ERROR: ";
	 std::cout << c << "\n";
	 return 0;
      }
   }
   else
   {
      try
      {
         readproblem( is, m, rs, fol );
      }
      catch( parser::syntaxerror s )
      {
         std::cout << "ERROR: "; 
         std::cout << s << "\n\n";
         return 0; 
      }
   }

   fol. collectusedidentifiers( );

   firstorder::simplifications::flatten( fol ); 

   if( fol. formulas. size( ))
   {

      firstorder::simplifications::flatten( fol );
      firstorder::nnf::convert_to_nnf( fol );
      firstorder::simplifications::flatten( fol ); 
      firstorder::simplifications::fully_simplify( fol );
      firstorder::antiskolemization::antiskolemize( fol );
      firstorder::simplifications::flatten( fol );
      firstorder::factor::convert_to_cuden( fol );
      firstorder::simplifications::flatten( fol );

      firstorder::geometric::addtherules( fol, rs );

   }

   analysis::dependency_system ds =
      analysis::dependency_system::computedependencies(rs); 

   {
      // We add the inductive axioms arising from the dependencies
      // to the rulesystem. 

      std::vector< predicate > predicates_base;
      std::vector< variable > arguments_base; 
    
      std::list< rule > 
         inductive = ds. axioms( predicates_base, arguments_base );

      std::vector< rulesystem::binary_step > deriv;

      for( std::list< rule > :: const_iterator 
              it = inductive. begin( );
              it != inductive. end( );
              ++ it )
      {
         rs. addrule( *it, deriv, true );
      }
   }


   if( pars. nonempty )
   {
      if( m. nr_elements( ) == 0 )
      {
         // In that case, we insert one element into the model:

         element e0;
	 std::vector< element > args;
	 args. push_back( e0 );

	 std::vector< unsigned int > parents;

         m. addatom( groundatom( predicate::create, args ), 9999999, parents );
      }
   }

   std::vector< geometric::rule > initials;
   std::vector< geometric::rule > simplifiers;

   for( unsigned int nr = 0; nr < rs. nr_rules( ); ++ nr )
   {
      if( ! rs. rules [nr]. deleted )
      {
         const rule& r = rs. rules[ nr ].r;

         if( r. isdisjunctive( ))
         {
            geometric::disjunctive disj; 
                for( auto p = r. prems_begin( ); p != r. prems_end( ); ++ p )
            {
               disj. insert_lhs( makenew( sig, *p, geometric::truthval:: F ));
            }

            for( auto p = r. concs_begin( ); p != r. concs_end( ); ++ p )
            {
               disj. insert_rhs( makenew( sig, *p, geometric::truthval:: T ));
            }
           
            if( rs. rules[nr]. inductive ) 
               simplifiers. push_back( { disj, geometric::unsound( ) } );
            else
               initials. push_back( { disj, geometric::input( ) } );
         }
         else
         {
            geometric::existential exists{ 
               makenew( *r. getcreate( ). a1 ), 
               makenew( sig, r. getbody( ), geometric::truthval::T ) };  
           
            for( auto p = r. prems_begin( ); p != r. prems_end( ); ++ p )
            {
               exists. insert_lhs( makenew( sig, *p, geometric::truthval:: F ));
            }

            if( rs. rules[nr]. inductive )
               simplifiers. push_back( { exists, geometric::unsound( ) } ); 
            else
               initials. push_back( { exists, geometric::input( ) } );
         }
      }

   }

   geometric::rulesystem sys{ "INPUT", initials, simplifiers };

   unsigned int rand = 0;
   while( true )
   {

      std::vector< geometric::groundatom > startint = { };
      if( pars. nonempty )
      {
         startint. push_back( 
              { sig. create, geometric::truthval::T, { e0 } } );
      }

      geometric::searchstate state(
      { geometric::permutationgenerator::strategy::random, 1 },
        geometric::rulesystem(sys), startint );
      
      try
      {
         bool b = state. search( 3600 * 24 * 7 );

         if( hasconjecture )
         {
            if(b)
               std::cout << "% SZS status CounterSatisfiable ";
            else
               std::cout << "% SZS status Theorem ";
         }
         else
         {
            if(b)
               std::cout << "% SZS status Satisfiable ";
            else
               std::cout << "% SZS status Unsatisfiable ";
         }
         std::cout << "for " << pars. inputfilename << "\n";

         std::cout << state << "\n";

         if(b)
         {
            std::cout << "% SZS output start Model for ";
            std::cout << pars. inputfilename << "\n\n";

            std::cout << state. intp << "\n";

            std::cout << "% SZS output end Model for ";
            std::cout << pars. inputfilename << "\n\n";
         }
         else
         {
            std::cout << "% SZS output start Refutation for ";
            std::cout << pars. inputfilename << "\n\n";

            std::cout << state. rs << "\n";

            std::cout << "% SZS output end Refutation for ";
            std::cout << pars. inputfilename << "\n\n";

         }

         return 0; 
      }
      catch( const std::runtime_error& e )
      {
         std::cout << "runtime error: " << e. what( ) << "\n";
         std::cout << "-------------------------\n";
         std::cout << "Timed out, and restarting with seed " << rand << ":\n";
         std::cout << "\n";
      }
   }

}


