
// Written by Hans de Nivelle, Dec 2015.
// Rewritten in May 2016.

#include "gcsp.h"

namespace geometric
{
namespace constraints
{

   gcsp::clause::state gcsp::clause::state::first( size_t s ) const
   {
      if( s0 + s > s1 )
         return *this; 
      else
         return state( s0, s0 + s );
   } 


   void gcsp::clause::state::next( const state& orig )
   {  
      size_t s = s1 - s0;
      s0 = s1;
      s1 = s1 + s;

      if( s1 > orig. s1 )
         s1 = orig. s1;
   }

 
   gcsp::clause::clause( const varatom& at )
      : initial{0,0},
        current{0,0} 
   {
      for( variable a : at )
      {
         for( auto p = domain. begin( ); p != domain. end( ); ++ p )
         {
            if( a == *p ) 
               goto dontadd;
         }
         domain. push_back(a);
      dontadd: 
         ;
      }
   }


   gcsp::clause::clause( const varatom& at, variable ex )
      : initial{0,0},
        current{0,0} 
   {
      for( variable a : at )
      {
         if( a == ex ) goto dontadd;
         for( auto p = domain. begin( ); p != domain. end( ); ++ p )
         {
            if( a == *p )
               goto dontadd;
         }
         domain. push_back(a);
      dontadd:
         ;
      }
   }


   bool gcsp::clause::indomain( variable v ) const
   {
      for( auto w : domain )  
      {
         if( v == w ) return true;
      }
      return false;
   }


   bool gcsp::clause::addconflict( const groundatom& atom1, 
                                   const varatom& atom2, 
                                   const set< size_t > * alpha )
   {
      geometric::subststack< element > subst;

      if( ( atom1.t & atom2.t ) == truthval::empty &&
            subst. match_arguments( atom2, atom1 ))
      {
         add_matching( subst.s, alpha );
         return true;
      }
      else
         return false;
   }


   bool gcsp::clause::addimplication( const groundatom& from,
                                      const varatom& into )
   {
      geometric::subststack< element > subst;

      if( ( from.t & into.t ) == from.t &&
            subst. match_arguments( into, from ))
      {
         add_matching( subst.s, nullptr );
         return true;
      }
      else
         return false;
   }


   bool gcsp::clause::addimplication( element e )
   {
      subst< element > subst;
      subst. assign( domain[0], e );
      subst. assign( domain[1], e );

      add_matching( subst, nullptr );
         // We don't need blocking because we have range restriction.

      return true;
   }


   void 
   gcsp::clause::set_current_from_alpha( bool bl, const set< size_t > * alpha )
   { 
      current = allmatchings( );

      size_t s = current. s0; 
      while( s < current. s1 )
      {
         bool included;

         if( matchings[s]. alpha )
         {
            if( alpha )
               included = subset( * matchings[s]. alpha, * alpha );
            else
               included = true;   // Absent == universal.
         }
         else
            included = ( alpha == nullptr ); 
               // If matchings[s]. alpha is universal, then alpha must be
               // universal too. 

         if( bl ? ( !matchings[s]. alpha || !included ) : included ) 
            ++ s; 
         else
            disable(s); 
      }

      initial = current;
   }


   gcsp::gcsp( const interpretation& intp, const indexedrule& from )
   {
      for( size_t nr = 0; nr != from. prem. size( ); ++ nr )
      {
         const varatom& atom = from. prem[nr];

         if( atom. p. isequality( ))
         {
            // Equalities are added to the avoids:

            blockings. push_back( clause( atom ));

            // We add all pairs ( E, E ) with unconditional assumption set,
            // because they would imply truth of the equality. 

            for( element e : intp. elements )
               blockings. back( ). addimplication(e);    

         }
         else
         {
            // Create the clause:

            clauses. push_back( clause( atom ));

            // List of occurrences of predicate from index:

            auto occ = intp. atoms. find( atom. p ); 

            for( auto p = occ. p1; p != occ. p2; ++ p )
               clauses. back( ). addconflict( intp. atoms[ *p ], atom,
                                              & intp. assumptions[ *p ] ); 
         }
      }

      if( from.r. isdisjunctive( ))
      {
         const disjunctive& disj = from.r. getdisjunctive( );

         for( size_t j = 0; j != disj. rhs_size( ); ++ j )
         {
            const varatom& rhs_atom = disj[j];

            // Create the blocking:

            blockings. push_back( clause( rhs_atom ));

            // Get a list of occurrences of the predicate from index:

            auto occ = intp. atoms. find( rhs_atom. p );

            for( auto p = occ. p1; p != occ. p2; ++ p )
            {
               blockings. back( ). addimplication( intp. atoms[ *p ], 
                                                   rhs_atom );
               blockings. back( ). addconflict( intp. atoms[ *p ], rhs_atom,
                                         & intp. assumptions[ *p ] );
            }
         }

      }
      else
      {
         const existential& ex = from.r. getexistential( );

         variable v = ex. getvariable( );
         const varatom& exatom = ex. getatom( );

         // Get a list of occurrences of the predicate from index:

         auto occ = intp. atoms. find( exatom. p );

         // Create the blocking:

         blockings. push_back( clause( exatom, v ));

         for( auto p = occ. p1; p != occ. p2; ++ p )
         {
            blockings. back( ). addimplication( intp. atoms[ *p ], exatom );
            blockings. back( ). addconflict( intp. atoms[ *p ], exatom,
                                      & intp. assumptions[ *p ] );

         }
      }

      build_index( ); 
   }   


   bool geometric::constraints::gcsp::occ_vect::contains( size_t k ) const
   {
      size_t lower = 0;
      size_t upper = vect. size( );

      while( upper - lower > 4 )
      {
         size_t middle = lower + ( upper - lower ) / 2;

         if( vect[ middle ] <= k )
            lower = middle;
         else
            upper = middle;
      }

      // End with linear search:

      while( lower < upper )
      {
         if( vect[ lower ] == k )
            return true;

         ++ lower;
      }

      return false;
   } 


   std::ostream& operator << ( std::ostream& out, const gcsp::occ_vect& occ )
   {
      out << "{";
      for( auto p = occ. vect. begin( ); p != occ. vect. end( ); ++ p )
      {
         if( p != occ. vect. begin( ))
            out << ", ";
         else
            out << " ";
         out << *p;
      }
      out << " }";
      return out; 
   }


   void geometric::constraints::gcsp::build_index( )
   {
      for( size_t c = 0; c < clauses. size( ); ++ c ) 
      {
         for( size_t v = 0; v < clauses[c]. nrvars( ); ++ v ) 
            varindex[ clauses[c]. var(v) ]. cl. append(c); 
      }

      for( size_t b = 0; b < blockings. size( ); ++ b )
      {
         for( size_t v = 0; v < blockings[b]. nrvars( ); ++ v ) 
           varindex[ blockings[b]. var(v) ]. bl. append(b); 
      }

   }


   std::pair< size_t, size_t >
   gcsp::getblocked( const subststack< size_t > & theta, size_t s ) const
   {
      while( s < theta. size( )) 
      {
         variable v = theta[s]. first;

         const auto& occ = varindex. at(v);

         for( size_t b : occ.bl )
         {
            // These are all the blockings that contain variable v.
            // The blocking must be fully instantiated, and
            // must have a choice that is consistent with theta. 

            if( instantiates( theta, blockings[b] ))
            {
               for( size_t j = blockings[b]. current. s0; 
                           j < blockings[b]. current. s1; ++ j )
               {
                  if( consistent( theta, blockings[b], j ))
                     return { b, j };
               } 
            }
         }
         ++ s;
      }
      return { blockings. size( ), 0 }; 
   }


   bool gcsp::impliesblocked( const subststack< size_t > & theta,
                              const clause& cl, size_t j ) const
   {
      // Enumerate all variables that occur in cl:

      for( size_t i = 0; i < cl. nrvars( ); ++ i )
      {
         variable v = cl. var(i); 
         const auto& occ = varindex. at( v );

         for( size_t b : occ.bl )
         {
            // We are now enumerating all blockings that share a variable 
            // with cl. 

            // We check if each variable in blockings[b] is assigned in 
            // theta, or occurs in cl.

            bool checkable = true; 
            for( size_t c = 0; c < blockings[b]. nrvars( ) && checkable; ++ c )
            {
               variable w = blockings[b]. var(c); 

               if( !theta. lookup(w) && !cl. indomain(w) )
                  checkable = false; 
            }

            if( checkable )
            {
               // As far as I can see, it is sufficient
               // for the c-th choice, to be consistent
               // with theta and with cl/j.

               for( size_t c = blockings[b]. current. s0; 
                           c < blockings[b]. current. s1; ++ c )
               {
                  if( consistent( theta, blockings[b], c ) &&
                      consistent( cl, j, blockings[b], c ))
                  {
                     return true;
                  }
               }
            }
         }
      }

      return false;
   }


   void gcsp::initfromalpha( const set< size_t > * alpha )
   {
      for( auto& cl : clauses )
         cl. set_current_from_alpha( false, alpha );

      for( auto& bl : blockings )
         bl. set_current_from_alpha( true, alpha );  
   }


   bool consistent( const subststack< size_t > & theta,
                    const gcsp::clause& cl, size_t j )
   {
      for( size_t v = 0; v != cl. nrvars( ); ++ v )
      {
         variable var = cl. var(v);
         size_t val = cl. value( v, j );

         const auto* p = theta. lookup( var );

         if( p && *p != val )
            return false;
      }
      return true;
   }


   bool consistent( const gcsp::clause& cl1, size_t j1, 
                    const gcsp::clause& cl2, size_t j2 )
   {
      for( size_t v1 = 0; v1 != cl1. nrvars( ); ++ v1 )
      {
         variable var1 = cl1. var( v1 );
         size_t val1 = cl1. value( v1, j1 );
        
         for( size_t v2 = 0; v2 != cl2. nrvars( ); ++ v2 )  
         {
            variable var2 = cl2. var( v2 );
            size_t val2 = cl2. value( v2, j2 );

            if( var1 == var2 )
               if( val1 != val2 ) return false;
            else
               goto found;   // Variables occur at most once.
         }
      found:
         ;  
      }
      return true;
   }


   bool match( subststack< size_t > & theta,
               const gcsp::clause& cl, size_t j )
   {
      for( size_t v = 0; v != cl. nrvars( ); ++ v )
      {
         variable var = cl. var(v);
         size_t val = cl. value( v, j );

         const auto* p = theta. lookup( var );
         if( !p )
            theta. assign( var, val );
         else
         {
            if( *p != val )
               return false;
         }
      }
      return true;
   }


   bool instantiates( const subststack< size_t > & theta,
                      const gcsp::clause& cl )
   {
      for( size_t v = 0; v != cl. nrvars( ); ++ v )
      {
         variable var = cl. var(v);

         // Every variable var in cl[i] must have assignment val: 

         const auto* p = theta. lookup( var );
         if( !p ) 
            return false;
      }
      return true;
   }


   bool gcsp::issolution( const subststack< size_t > & theta, 
                          bool tellreason ) const
   {
      for( size_t c = 0; c < clauses. size( ); ++ c )
      {
         if( !instantiates( theta, clauses[c] ))
         {
            if( tellreason )
               std::cout << "clause " << c << " is not fully instantiated\n";

            return false;
         }

         for( size_t j = clauses[c]. current. s0; 
                     j < clauses[c]. current. s1; ++ j )
         {
            if( consistent( theta, clauses[c], j ))
               goto found;
         }

         if( tellreason )
            std::cout << "clause " << c << " is not true\n";

         return false;  
      found:  
         ;
      }

      for( size_t b = 0; b < blockings. size( ); ++ b )
      {
         if( !instantiates( theta, blockings[b] ))
         {
            if( tellreason )
               std::cout << "blocking " << b << " is not fully instantiated\n"; 

            return false;
         }
      
         for( size_t j = blockings[b]. current. s0; 
                     j < blockings[b]. current. s1; ++ j )
         {
            if( consistent( theta, blockings[b], j ))
            {
               if( tellreason )
                  std::cout << "blocking " << b << "/" << j << " is true\n"; 

               return false; 
            }
         }
      }

      return true;
   }


   set< size_t >
   gcsp::getassumptions( const subststack< size_t > & theta ) const
   {
      set< size_t > alpha;

      for( size_t c = 0; c < clauses. size( ); ++ c )
      {
         for( size_t j = clauses[c]. current. s0; 
                     j < clauses[c]. current. s1; ++ j )
         {
            if( consistent( theta, clauses[c], j ))
               alpha |= clauses[c]. getalpha(j); 
         }
      }

      for( size_t b = 0; b < blockings. size( ); ++ b )
      {
         // The following probably requires some explanation: Blockings
         // in the current state are certainly not true. This follows from 
         // the fact that theta is a solution. Hence there is no point in 
         // looking at them. Choices outside our state are choices that 
         // would create a conflict that we currently allow. Even though 
         // we allow the conflict, we still have to include it. 

         clause::state all = blockings[b]. allmatchings( );
         clause::state current = blockings[b]. current;
     
         for( size_t j = all. s0; j < all. s1; ++ j )
         {
            if( !current. contains(j) && consistent( theta, blockings[b], j ))
            {
               alpha |= blockings[b]. getalpha(j);
            }

            // It follows from the fact that the clause was unusable,
            // that it carries an alpha. 
         }
      }

      return alpha; 
   }

   
   std::ostream& operator << ( std::ostream& out,
                               const gcsp::occurrences& occ )
   {
      out << "in clauses: " << occ. cl << ",      ";
      out << "in blockings: " << occ. bl;
 
      return out;
   }


   std::ostream& operator << ( std::ostream& out, const gcsp::matching& m )
   {
      out << "(";
      for( auto p = m. values. begin( ); p != m. values. end( ); ++ p )
      {
         if( p == m. values. begin( ))
            out << " ";
         else
            out << ", ";
         out << *p;
      }
      out << " )";

      if( m. alpha )
         out << "/" << *( m. alpha );

      return out;
   }


   std::ostream& 
   operator << ( std::ostream& out, const gcsp::clause::state s )
   {
      out << "[ " << s. s0 << " .. " << s. s1 << " )";
      return out;
   }


   void gcsp::clause::hard_commit( )
   {
      matchings. erase( matchings. begin( ) + current. s1, matchings. end( ));

      size_t s = current. s0;
      matchings. erase( matchings. begin( ), matchings. begin( ) + s );

      current. s0 -= s;
      current. s1 -= s;

      initial. s0 -= s;
      initial. s1 -= s;
   }


   void gcsp::clause::print( std::ostream& out ) const 
   {
      out << "(";

      for( size_t v = 0; v < nrvars( ); ++ v )
      {
         if( v == 0 ) 
            out << " ";
         else
            out << ", ";
         out << var(v); 
      }
      out << " ) ==> ";

      for( size_t j = current. s0; j < current. s1; ++ j ) 
      {
         if( j != 0 )
            out << " | ";
   
         out << j << ":" << matchings[j];
      }
   }


   std::ostream& operator << ( std::ostream& out, const gcsp& g )
   {
      out << "Generalized Constrained Solving Problem:\n";
      out << "----------------------------------------\n";
      out << "\n";

      out << "Clauses:\n";
      for( size_t nr = 0; nr < g. clauses. size( ); ++ nr )
         out << nr << " : " << g. clauses[ nr ] << "\n";
      out << "\n";

      out << "Blockings:\n";
      for( size_t nr = 0; nr < g. blockings. size( ); ++ nr )
         out << nr << " : " << g. blockings[ nr ] << "\n";
      out << "\n";

      out << "Index:\n";
      for( const auto& c : g. varindex )
      {
         out << "   ";
         out << c.first << ": ";
         out << c. second << "\n";
      } 
      return out;
   }


}}




