
// Written by Hans de Nivelle, December 2015.
// Rewritten in May 2016:
//    I added backtracking support to gcsp::clause, and
//    removed class choicestack.

#ifndef GEOMETRIC_CONSTRAINTS_GCSP_INCLUDED
#define GEOMETRIC_CONSTRAINTS_GCSP_INCLUDED  1

#include <vector>
#include <unordered_map>

#include "../set.h"
#include "../interpretation.h"
#include "../rulesystem.h"
#include "subststack.h"


namespace geometric
{
namespace constraints 
{

   struct gcsp
   {

      struct matching
      {
         std::vector< size_t > values; 
         const set< size_t > * alpha;
            // We are not owner! The alphas are probably part of an 
            // interpretation. Blockings that originate from a true 
            // conclusion have no alpha. 

         matching( const set< size_t > * alpha )
            : alpha{ alpha }
         { }
      };


      class clause
      {
         std::vector< variable > domain;     
         std::vector< matching > matchings;

      public: 
         struct state
         {
            size_t s0;
            size_t s1;

            state( size_t s0, size_t s1 )
               : s0{s0}, s1{s1}
            { }

            bool contains( size_t s ) const { return s >= s0 && s < s1; } 
            bool included_in( state s ) const 
               { return s0 >= s.s0 && s1 <= s.s1; }

            size_t size( ) const { return s1 - s0; } 

            state first( size_t s ) const;  // Choose first s substlets.
            void next( const state& orig ); 
               // Increase *this, while making sure that we stay within orig.
               // If we reach the end of orig, the new state will be empty.
         };

         state initial;   // State that is used for lemma derivation.
                          // Initial is not used in blockings.
         state current;   // Current state during search. 
            // All search algorithms must have invariant that no
            // matchings outside the current interval. 
            // Also, it seems reasonable that current is a subinterval of
            // initial. 
            // Exchanging matchings inside current is possible. 

         clause( const varatom& at );
            // Make a clause whose domain consists of the
            // variables in at. The clause has no matchings.
   
         clause( const varatom& at, variable ex );
            // Make a clause whose domain consists of the 
            // variables occurring in at without ex. The clause has 
            // no matchings.
            // (It is assumed that ex is existentially quantified.) 

         // Type T must have a field getindex( ).

         template< typename T > 
         void add_matching( const subst<T> & s,
                            const set< size_t > * alpha ) 
         {
            matchings. push_back( matching( alpha ));
            for( variable v : domain )
               matchings. back( ). values. push_back( 
                                      s. clookup(v) -> getindex( ));
         } 

         size_t nrvars( ) const { return domain. size( ); }
         variable var( size_t i ) const { return domain[i]; }

         bool indomain( variable v ) const;
           // True if variable v is in our domain.

         size_t value( size_t v, size_t j ) const
            { return matchings[j]. values[v]; }
               // Value for the v-th variable in the j-th choice.

         bool hasalpha( size_t j ) const
            { return matchings[j]. alpha; }

         const set< size_t > getalpha( size_t j ) const
            { return *matchings[j]. alpha; } 

         bool addconflict( const groundatom& atom1, const varatom& atom2,
                           const set< size_t > * alpha );
            // If atom2 has an instance that conflicts atom1, then 
            // add the vector of values for the variables that occur in 
            // our domain. For the remaining variables, we assume that
            // they were existentially quantified, and should therefore 
            // not be mentioned. Conflicts have an assumption set, 
            // originating from the conflict set of atom2 in the 
            // interpretation.

         bool addimplication( const groundatom& from, const varatom& into );
            // If from implies an instantiation of into, then
            // add the values of the variables that are in the domain to
            // the clause. Blocks caused by the truth of a conclusion
            // have no assumption set.

         bool addimplication( element e );
            // Assuming that the clause is derived from an equality,
            // add (e,e), denoting that (X,Y) := (e,e) makes the
            // equality true. 

         void disable( size_t i )
         {
            -- current. s1;
            if( i != current. s1 )
               std::swap( matchings[i], matchings[ current. s1 ] );
         }
            // Disable the i-th choice, by swapping it with the last
            // element in current. This will decrease current.s1 by one.
            // If you want to check many choices, do:
            // i = current.i0; while( i < current.i1 ) 
            //    if( i is OK ) i++; else disable(i); 
            // If you have a list of element to delete, you must
            // do it backwards:
            // i = current.i1; while( i != current.i0 ) 
            //    --i;  disable(i);
        
         void set_current_from_alpha( bool bl, const set< size_t > * alpha );
            // Set the current and the initial state using alpha. A nullptr 
            // is treated as a universal alpha (containing everything).
            // If we are not a blocking, the alpha of the matching
            // must be included in the parameter alpha. 
            // If we are a blocking, we must have empty alpha, or our
            // alpha must be not included in the parameter alpha. 
            // This is Definition 20 in the paper.
            // Note that we use disable( ), so that matchings can be  
            // permuted. 

         state allmatchings( ) const { return { 0, matchings. size( ) }; }
            // All matchings in the clause. 

         void hard_commit( );
            // Permanently remove all choices that are not current.

         void print( std::ostream& ) const; 

      };



      std::vector< clause > clauses;
      std::vector< clause > blockings;
         // Blockings are not clauses in strict sense because they   
         // do not have disjunctive meaning. It is still convenient
         // to represent them as clauses because avoids originating
         // from the same conclusion have the same set of variables. 
         // 
         // If a value in a blocking has no assumption set, 
         // then the blocking should be always used.
         // If a value has an assumption set, then the blocking should 
         // be used when its assumption set is not included in the current 
         // assumption set. 

      // Increasing vector of occurrences of a variable.

      struct occ_vect
      {
         std::vector< size_t > vect;

         occ_vect( ) { } 

         bool contains( size_t k ) const;

         void append( size_t k )
         {
            if( vect. size( ) == 0 || vect. back( ) != k )
               vect. push_back(k);
         }


         // Only const iteration:

         std::vector< size_t > :: const_iterator begin( ) const
            { return vect. cbegin( ); }
         std::vector< size_t > :: const_iterator end( ) const
            { return vect. cend( ); } 
      };
   

      // List of occurrrences of a variable:
      //   cl : Clauses in which the variable occurs.
      //   bl : Blockings in which the variable occurs

      struct occurrences
      {
         occ_vect cl;
         occ_vect bl;
      };


      std::unordered_map< variable, occurrences,
                          variable::hash, variable::equal_to > varindex;
         // An index to the occurrences of a variable in
         // clauses and blockings.


      gcsp( const interpretation& intp, const indexedrule& from );
         // Rule from can be existential or disjunctive.

      void build_index( );
         // Build the index.

      const occ_vect& inclauses( variable v ) const
         { return varindex. at(v). cl; }
      const occ_vect& inblockings( variable v ) const
         { return varindex. at(v). bl; }

      std::pair< size_t, size_t >
      getblocked( const subststack< size_t > & theta, size_t s ) const;
         // If theta (involving a recent change >= s ) 
         // implies a blocking, we return such a blocking.
         // Otherwise, we return ( blockings. size( ), 0 ). 
 
      bool impliesblocked( const subststack< size_t > & theta, size_t s ) const
         { return getblocked( theta, s ). first < blockings. size( ); }

         // True if one of the recent assignments (>= s),
         // possibly together with older assignments implies a 
         // blocked clause. There must be a recent assignment involved, 
         // in order to avoid rechecks during backtracking.

      bool impliesblocked( const subststack< size_t > & theta,
                           const clause& cl, size_t j ) const;
         // True if theta combined with cl/j implies a blocking.
         // Note that we assume that theta and cl/j are consistent.
         // Consistency between theta and cl/j must be checked separately.

      void initfromalpha( const set< size_t > * alpha ); 
         // Using alpha, initialize the current states in the 
         // clauses and in the blockings. Note that initial states
         // are not initialized. Use set_initials( ) for this. 

      bool issolution( const subststack< size_t > & theta,
                       bool tellreason = false ) const;
         // True if theta is a solution. If tellreason is true,
         // we print the reason why it is not a solution, in case
         // of returning false. 

      set< size_t > 
      getassumptions( const subststack< size_t > & theta ) const;
         // Get the assumption set of a given theta.
         // It is obtained from the usable clauses that are made true
         // by theta, and the unusable blockings that are made true. 
         // theta must be a solution! 

      void set_initials( )
         { for( auto& cl : clauses ) cl. initial = cl. current; }
            // Set the initial states of the clauses equal to the current
            // states. This should be done after preprocessing. 
      
      void hard_commit( ) 
         { for( auto& cl : clauses ) cl. hard_commit( ); } 
         // Permantenly remove all substlets that are not current
         // from the clauses. Blockings are not touched. 
         // initial is adopted. This method moves substlets.
   };


   bool consistent( const subststack< size_t > & theta,
                    const gcsp::clause& cl, size_t j );
      // True if subst is consistent with cl/j.
 
   bool consistent( const gcsp::clause& cl1, size_t j1, 
                    const gcsp::clause& cl2, size_t j2 );
      // True if cl1/j1 and cl2/j2 do not have conflicting assignments.

   bool match( subststack< size_t > & theta,
               const gcsp::clause& cl, size_t j );
      // Attempt to extend theta in such a way that it implies cl/j.
      // Note that theta may still be extended in case of failure!

   bool instantiates( const subststack< size_t > & theta, 
                      const gcsp::clause& cl );
      // True if subst instantiates all variables in cl.

   std::ostream& operator << ( std::ostream& out, const gcsp::matching& m );
   std::ostream& operator << ( std::ostream& out, const gcsp::clause::state );

   inline std::ostream& 
   operator << ( std::ostream& out, const gcsp::clause& cl )
   {
      cl. print( out );
      return out;
   }

   std::ostream& operator << ( std::ostream& out, const gcsp::occ_vect& occ );

   std::ostream& operator << ( std::ostream& out, 
                               const gcsp::occurrences& occ ); 
   std::ostream& operator << ( std::ostream& out, const gcsp& g );

}}

#endif

