
#include "lemma.h"

namespace geometric
{
namespace constraints
{

   void lemma::insert( const gcsp::clause& cl, size_t i )   
   {
      size_t base = assignments. size( );
      for( size_t v = 0; v < cl. nrvars( ); ++ v )
         assignments. push_back( { cl. var(v), cl. value( v, i ) } );

      substlets. push_back( { base, assignments. size( ) } );
      check_last_subsumption( );
   }

   void 
   lemma::insert( 
      const std::vector< std::pair< const gcsp::clause*, size_t >> & vect )
   {
      size_t base = assignments. size( );
      
      for( const auto& p : vect )
      {
         const gcsp::clause* cl = p. first; 

         for( size_t v = 0; v < cl -> nrvars( ); ++ v )
         {
            variable var = cl -> var( v );
            size_t val = cl -> value( v, p. second );

            for( size_t a = base; a < assignments. size( ); ++ a )
            {
               if( assignments[a]. var == var )
               {
                  if( assignments[a]. val == val )
                     goto found_agreed;
                 
                  // Not agreed means that we don't add anything.

                  while( assignments. size( ) > base )
                     assignments. pop_back( );
                  return; 
               }
            }
            assignments. push_back( { var, val } );
         found_agreed:
            ;
         }
      }

      substlets. push_back( { base, assignments. size( ) } );
      check_last_subsumption( );
   }


   void lemma::insert( const lemma& lem, size_t i )
   {
      size_t base = assignments. size( ); 
      for( size_t a = lem. substlets[i]. a0; a < lem. substlets[i]. a1; ++ a )
         assignments. push_back( lem. assignments[a] );
      substlets. push_back( { base, assignments. size( ) } );
      check_last_subsumption( );
   } 

   void 
   lemma::insert( std::initializer_list< std::pair< variable, size_t >> ins )
   {
      size_t base = assignments. size( );
      for( auto a : ins )
         assignments. push_back( { a. first, a. second } );
      substlets. push_back( { base, assignments. size( ) } );
      check_last_subsumption( );
   }


   bool 
   geometric::constraints::lemma::implies( 
      const lemma& lem1, size_t i1,
      const lemma& lem2, size_t i2 )
   {
      // Every assignment in lem2[i2] must occur in lem1[i1]:

      for( size_t a2 =  lem2. substlets[i2]. a0;
                  a2 != lem2. substlets[i2]. a1;
                  ++ a2 )
      {
         bool found = false;

         for( size_t a1 =  lem1. substlets[i1]. a0;
                     a1 != lem1. substlets[i1]. a1 && !found;
                     ++ a1 )
         {
            if( lem1. assignments[a1]. var == lem2. assignments[a2]. var &&
                lem1. assignments[a1]. val == lem2. assignments[a2]. val )
            {
               found = true; 
            }
         }
         if( !found )
            return false; 
      }
      return true;
   }


   void lemma::check_last_subsumption( )
   {  
      for( size_t i = 0; i + 1 < substlets. size( ); ++ i )
      {
         if( implies( *this, i, *this, substlets. size( ) - 1 ))
         {
            remove( substlets. size( ) - 1 );
            return;
         }
      }

      size_t i = 0; 
      while( i + 1 < substlets. size( ))
      {
         if( implies( *this, substlets. size( ) - 1, *this, i ))
         {
            remove(i);   // This is safe, because only indices exist.
         }
         else
            ++ i;
      }
   }


   void lemma::remove( size_t i )
   {
      size_t diff = substlets[i]. a1 - substlets[i]. a0;
      assignments. erase( assignments. begin( ) + substlets[i]. a0,
                          assignments. begin( ) + substlets[i]. a1 );
      substlets. erase( substlets. begin( ) + i );

      // Since substlets are implemented as references into assignments, 
      // they have to be corrected:

      while( i < substlets. size( ))
      {
         substlets[i]. a0 -= diff;
         substlets[i]. a1 -= diff;
         ++ i;
      }
   }


   std::ostream& operator << ( std::ostream& out, const lemma::assignment& a )
   {
      out << a. var << "/" << a. val;
      return out; 
   }


   std::ostream& operator << ( std::ostream& out, const lemma& lem )
   {
      out << "lemma{";
      for( size_t i = 0; i < lem. nrsubstlets( ); ++ i )
      {
         if(i)
            out << ", ";
         else
            out << " ";

         out << "(";
         for( size_t a = 0; a < lem. nrassignments(i); ++ a )
         {
            if(a) out << ",";
            out << lem. getassignment(i,a). var;
         }

         out << ")/(";
         for( size_t a = 0; a < lem. nrassignments(i); ++ a )
         {
            if(a) out << ",";
            out << lem. getassignment(i,a). val; 
         }
         out << ")"; 
      }
      out << " }";
      return out; 
   }


   std::ostream& operator << ( std::ostream& out, const confl_set& c )   
   {
      out << "{";
      for( auto p = c. begin( ); p != c. end( ); ++ p )
      { 
         if( p != c. begin( ))
            out << ", ";
         else
            out << " ";
 
         out << *p;
      }
      out << " }";
      return out;
   }


   bool consistent( const subststack< size_t > & theta, 
                    const lemma& lem, size_t i )
   {
      // std::cout << "consistent " << lem << "/" << i << "\n";

      for( size_t a = 0; a != lem. nrassignments(i); ++ a )
      {
         auto assign = lem. getassignment(i,a);

         const auto* p = theta. lookup( assign. var );
         if( p && *p != assign. val )
            return false;
      }
      return true;
   }


   bool consistent( const lemma& lem1, size_t i1,
                    const lemma& lem2, size_t i2 )
   {
      // std::cout << "consistent " << lem1 << "/" << i1 << "\n";
      // std::cout << "           " << lem2 << "/" << i2 << "\n";

      for( size_t a1 = 0; a1 != lem1. nrassignments( i1 ); ++ a1 )
      {
         for( size_t a2 = 0; a2 != lem2. nrassignments( i2 ); ++ a2 )
         {
            auto av1 = lem1. getassignment(i1,a1);
            auto av2 = lem2. getassignment(i2,a2);

            if( av1. var == av2. var && av1. val != av2. val )
            {
               return false;  
            }
         }
      }
      return true;
   } 


   bool
   consistent( const lemma& lem, size_t i, const gcsp::clause& cl, size_t j )
   {
      // std::cout << "consistent " << lem << " / " << i << "\n";
      // std::cout << "           " << cl << " / " << j << "\n";

      for( size_t a = 0; a != lem. nrassignments( i ); ++ a )
      {
         for( size_t v = 0; v != cl. nrvars( ); ++ v )
         {
            auto av = lem. getassignment(i,a); 
            if( av. var == cl. var(v) && av. val != cl. value(v,j) )
            {
               return false;
            }
         }
      }  

      return true;
   }


   bool consistent( const gcsp::clause& cl, const lemma& lem, size_t i )
   {
      for( size_t c = cl. current. s0; c != cl. current. s1; ++ c )
      {
         if( consistent( lem, i, cl, c ))
            return true;
      }
      return false;
   }


   bool consistent( const gcsp& g, const lemma& lem, size_t i )
   {
      for( size_t a = 0; a < lem. nrassignments(i); ++ a )
      {
         auto assign = lem. getassignment(i,a);
         const auto& occ = g. inclauses( assign. var );

         // Enumerating the clauses that share a variable with lem[i]:

         for( size_t sharing : occ )
         {
            if( !consistent( g. clauses[ sharing ], lem, i ))
               return false;
         }
      }
      return true; 
   }



   bool match( subststack< size_t > & theta, const lemma& lem, size_t i )
   {
      for( size_t a = 0; a != lem. nrassignments(i); ++ a ) 
      {
         auto assign = lem. getassignment(i,a);  

         const auto* p = theta. lookup( assign. var );
         if( !p )
            theta. assign( assign. var, assign. val );
         else
         {
            if( *p != assign. val )
               return false;
         }
      }
      return true;
   }


   bool istrue( const subststack< size_t > & theta, const lemma& lem )
   {
      for( size_t i = 0; i < lem. nrsubstlets( ); ++ i )
      {
         for( size_t a = 0; a != lem. nrassignments(i); ++ a )
         {
            auto assign = lem. getassignment(i,a);

            const auto* p = theta. lookup( assign. var );
            if( !p || *p != assign. val )
               goto nottrue;
         }
         return true;

      nottrue:
         ; 
      }
      return false;
   }


}}


