
// Written by Hans de Nivelle, April 2016.
// A lemma, as described in our IJCAR 2016 paper.

#ifndef GEOMETRIC_CONSTRAINTS_LEMMA_INCLUDED
#define GEOMETRIC_CONSTRAINTS_LEMMA_INCLUDED  1

#include <vector>
#include <iostream>

#include "subststack.h"
#include "gcsp.h"

namespace geometric 
{
namespace constraints 
{

   struct lemma 
   {

      struct assignment
      {
         variable var;
         size_t val; 

         assignment( variable var, size_t val )
            : var{ var }, val{ val }
         { } 
      };


   private: 
      std::vector< assignment > assignments;
          
      struct substlet
      {
         size_t a0;   // These are indices into assignments. 
         size_t a1;
      }; 

      std::vector< substlet > substlets;
      
   public: 
      lemma( ) { };

      lemma( const gcsp::clause& cl )
         { for( size_t i = cl. initial. s0; i < cl. initial. s1; ++ i ) 
             insert( cl, i ); 
         }

      lemma( std::initializer_list< std::initializer_list< 
                                    std::pair< variable, size_t >>> init )
         { for( auto s : init )
              insert( s );
         }


      void insert( const gcsp::clause& cl, size_t i );
      void insert( const std::vector< 
                      std::pair< const gcsp::clause*, size_t >> & vect );
      void insert( const lemma& lem, size_t i );
      void insert( std::initializer_list< std::pair< variable, size_t >> ins );
         // Insert cl[i], Prod( vect ), lem[i] or ins. Insertion involves
         // removal of implied literals, both backward
         // and forward. 
         // The product of vect equals the union, when vect is pairwise
         // consistent. 

      size_t nrsubstlets( ) const { return substlets. size( ); }
      size_t nrassignments( size_t i ) const 
         { return substlets[i]. a1 - substlets[i]. a0; } 

      assignment getassignment( size_t i, size_t a ) const
         { return assignments [ substlets[i]. a0 + a ]; } 

      static bool implies( const lemma& lem1, size_t i1,
                           const lemma& lem2, size_t i2 );
         // True if lem1[i1] implies lem2[i2].

   private:
      void check_last_subsumption( );
         // Remove the last substlet, if it is subsumed by earlier substlets.
         // If after that, the last substlets subsumes earlier substlets, 
         // then remove these substlets.

      void remove( size_t i );
         // Remove the i-th substlet.
   };


   std::ostream& operator << ( std::ostream& out, const lemma::assignment& );
   std::ostream& operator << ( std::ostream& out, const lemma& lem );


   bool consistent( const subststack< size_t > & theta,
                    const lemma& lem, size_t i );
      // True if theta and lem[i] are consistent.

   bool consistent( const lemma& lem1, size_t i1,
                    const lemma& lem2, size_t i2 );
      // True if lem1[i1] and lem2[i2] are consistent.

   bool consistent( const lemma& lem, size_t i,
                    const gcsp::clause& cl, size_t j );
      // True if lem[i] and cl[j] are consistent.

   bool consistent( const gcsp::clause& cl, const lemma& lem, size_t i );
      // True if cl is consistent with lem[i].
  
   bool consistent( const gcsp& g, const lemma& lem, size_t i ); 
      // True if g is consistent with lem[i].

   bool match( subststack< size_t > & theta, 
               const lemma& lem, size_t i );
      // Attempt to extend theta in such a way that it implies lem/i.
      // Note that theta may still be extended in case of failure!

   bool istrue( const subststack< size_t > & theta, const lemma& lem );
      // True if theta makes lem true. 

   // A conflict set represents a set of premisses of a lemma or clause,
   // with the aim of resolving them away. 
   // We assume that such sets are not very big.

   class confl_set
   {
      std::vector< size_t > prem;

   public: 
      confl_set( ) { };

      confl_set( const std::vector< size_t > & prem )
         : prem{ prem } 
      { }

      confl_set( std::vector< size_t > && prem )
         : prem{ std::move( prem ) } 
      { }

      confl_set( const confl_set& ) = default;
      confl_set( confl_set&& ) = default;
      confl_set& operator = ( const confl_set& ) = default;
      confl_set& operator = ( confl_set&& ) = default;

      size_t size( ) const 
         { return prem. size( ); }

      bool contains( size_t v ) const 
      {
         for( auto p : prem ) 
            if( p == v ) return true;
         return false;
      } 

      void insert( size_t v ) 
      {
         if( !contains(v))
            prem. push_back(v);
      }


      using iterator = std::vector< size_t > :: const_iterator;
         // We only allow const iteration. 

      iterator begin( ) const { return prem. cbegin( ); }
      iterator end( ) const { return prem. cend( ); }

      iterator cbegin( ) const { return prem. cbegin( ); }  
      iterator cend( ) const { return prem. cend( ); }
       
   };

   std::ostream& operator << ( std::ostream& out, const confl_set& s );

}}

#endif


