
// Written by Hans de Nivelle, May 2016.

#include "lemmasystem.h"
#include "resolution.h"
#include <stdexcept>


namespace geometric
{
namespace constraints
{

   size_t lemmasystem::insert( const lemma& lem, const gcsp& g )
   {
      size_t s = sys. size( );
      sys. push_back( lem ); 

      size_t i = firstconsistent( 
         [ &g ]( const lemma& lem, size_t i )
                  { return consistent( g, lem, i ); }, lem );

      if( i == sys. back( ). nrsubstlets( ))
         throw std::runtime_error( "insert: Lemma is in conflict" ); 
      addwatch( s, i );
      return s;
   }

   size_t lemmasystem::insert( lemma&& lem, const gcsp& g )
   {
      size_t s = sys. size( );
      sys. push_back( std::move(lem) );

      size_t i = firstconsistent( 
         [ &g ]( const lemma& lem, size_t i )
                  { return consistent( g, lem, i ); }, lem );

      if( i == sys. back( ). nrsubstlets( ))
         throw std::runtime_error( "insert: Lemma is in conflict" );
      addwatch( s, i );
      return s; 
   }

   size_t lemmasystem::insert( const lemma& lem, 
                               const subststack< size_t > & theta ) 
   {
      size_t s = sys. size( );
      sys. push_back( lem );

      size_t i = firstconsistent(
          [ &theta ]( const lemma& lem, size_t i )
               { return consistent( theta, lem, i ); }, lem );

      if( i == sys. back( ). nrsubstlets( ))
         throw std::runtime_error( "insert: Lemma is in conflict" );
      addwatch( s, i );
      return s;
   }

   size_t lemmasystem::insert( lemma&& lem, 
                               const subststack< size_t > & theta ) 
   {
      size_t s = sys. size( );
      sys. push_back( std::move(lem) );

      size_t i = firstconsistent(
          [ &theta ]( const lemma& lem, size_t i )
               { return consistent( theta, lem, i ); }, lem );

      if( i == sys. back( ). nrsubstlets( ))
         throw std::runtime_error( "insert: Lemma is in conflict" );
      addwatch( s, i );
      return s;
   }

   size_t
   lemmasystem::firstconsistent( 
         const std::function< bool( const lemma& , size_t ) > & consistency,
         const lemma& lem )
   {
      size_t c = 0;
      while( c < lem. nrsubstlets( ) && !consistency( lem, c ))
      {
         ++ c;
      }
      return c;
   }

   void 
   lemmasystem::checkwatching( variable v, size_t& current_confl, 
      const std::function< bool( const lemma& , size_t ) > & consistency )
   {
      // std::cout << "checkwatching " << v << ", " << current_confl << "\n";

      const auto& wl = watchlist[v];
      auto p = wl. begin( );
      while( p != wl. end( ))
      {
         // We are now enumerating the lemmas that watch variable v:

         size_t lem = p -> lem;
         if( !consistency( sys[ lem ], p -> i ))
         {
            size_t i2 = firstconsistent( consistency, sys[ lem ] );

            if( i2 < sys [ lem ]. nrsubstlets( ))
            {
               // We found another substlet that we can watch:

               removewatch( lem, p -> i );
                  // Not only from current watchlist, but from all lists.
               addwatch( lem, i2 );
                  // Note that we are playing with fire. The reinsertion
                  // of lem/i2 into our current list won't cause
                  // vector reallocation in wl, because we just deleted
                  // *p from it.

                  // p is not increased, because *p was removed.
            }
            else
            {
               // lem is inconsistent. We check if it shorter than 
               // current_confl:

               if( current_confl == size( ) ||
                  sys[ lem ]. nrsubstlets( ) < 
                     sys[ current_confl ]. nrsubstlets( ))
               {
                  current_confl = lem;
               }

               ++ p;
            }
         }
         else
            ++ p;
      }
   }

   size_t 
   lemmasystem::findconflict( const gcsp& g,
                              const std::vector< refinement > & stack,  
                              size_t k )
   {
      size_t confl = size( );

      while( k < stack. size( ))
      {
         auto const& cl = g. clauses[ stack[k]. cl ];
         for( size_t i = 0; i != cl. nrvars( ); ++ i )
         {
            variable v = cl. var(i);
            checkwatching( v, confl,
               [ &g ]( const lemma& lem, size_t i )
                  { return consistent( g, lem, i ); } );
         }

         ++ k;
      }

      return confl;
   }

   size_t 
   lemmasystem::findconflict( const subststack< size_t > & theta, size_t s )
   {
      size_t confl = size( );

      while( s < theta. size( ))
      {
         variable v = theta[s]. first; 

         checkwatching( v, confl,
            [ &theta ]( const lemma& lem, size_t i )
               { return consistent( theta, lem, i ); } );

         ++ s;
      }

      return confl;
   }



   void lemmasystem::addwatch( size_t lem, size_t i )
   {
      for( size_t a = 0; a < sys[ lem ].nrassignments(i); ++ a )
      {
         auto assign = sys[ lem ]. getassignment(i,a);
         insert( watchlist[assign.var], watched( lem, i ));
      }
   }

   void lemmasystem::removewatch( size_t lem, size_t i )
   {
      for( size_t a = 0; a < sys[ lem ].nrassignments(i); ++ a )
      {
         auto assign = sys[lem]. getassignment(i,a);
         remove( watchlist[assign.var], watched( lem, i ));
      }
   }

   void lemmasystem::insert( std::vector< watched > & watch, watched w )
   {
      watch. push_back(w);
   }


   
void lemmasystem::remove( std::vector< watched > & watch, watched w )
{
   auto p = watch. begin( );
   while( p != watch. end( ))
   {
      if( p -> lem == w. lem && p -> i == w. i )
      {
         watch. erase(p);
         return;
      }
      ++ p;
   }
}


std::ostream&
operator << ( std::ostream& out,
              const std::vector< lemmasystem::watched > & watchlist )
{
   out << "{";
   for( auto p = watchlist. begin( ); p != watchlist. end( ); ++ p )
   {
      if( p != watchlist. begin( ))
         out << ",";
      out << *p;
   }
   out << "}";
   return out;
}


std::ostream& operator << ( std::ostream& out, const lemmasystem& sys )
{
   out << "Lemma System:\n";

   for( size_t i = 0; i < sys. size( ); ++ i )
   {
      out << i << " : " << sys. sys[i] << "\n";
   }
   out << "\n";

   out << "Watching: \n";
   for( const auto& v: sys. watchlist )
      out << "   " << v. first << " -> " << v. second << "\n";

   return out;
}


}}



