
// Written by Hans de Nivelle, March-May 2016.
// Matching with positive learning, as described in the IJCAR 2016 paper.

#ifndef GEOMETRIC_CONSTRAINTS_LEMMASYSTEM_INCLUDED
#define GEOMETRIC_CONSTRAINTS_LEMMASYSTEM_INCLUDED  1

#include <vector>
#include <unordered_map>

#include "lemma.h"
#include "subststack.h"
#include "refinement.h"

namespace geometric 
{
namespace constraints 
{

   struct lemmasystem
   {

      // Meaning is that we are watching lem[i]. 

      struct watched
      {
         size_t lem;
         size_t i;

         watched( size_t lem, size_t i )
            : lem{ lem }, i{i}
         { }

      };


      std::vector< lemma > sys;

      lemmasystem( ) { } 

      size_t insert( const lemma& lem, const gcsp& g ); 
      size_t insert( const lemma& lem, const subststack< size_t > & theta ); 
      size_t insert( lemma&& lem, const gcsp& g );
      size_t insert( lemma&& lem, const subststack< size_t > & theta ); 
         // Lemmas can be inserted only when they are not a conflict lemma.
         // Otherwise, we throw unpleasant runtime errors.

      size_t findconflict( const gcsp& g, 
                           const std::vector< refinement > & stack, size_t k );
      size_t findconflict( const subststack< size_t > & theta, size_t s ); 
         // Returns shortest conflict lemma, or size( ) if there exists none.
 
      size_t size( ) const { return sys. size( ); }

      const lemma& operator [] ( const size_t i ) const
         { return sys[i]; }

      std::unordered_map< variable, std::vector< watched >,
                          variable::hash, variable::equal_to > watchlist;
         // Of a given watched substlet, all variables
         // are in the watchlist. 

      void clear( ) { sys. clear( ); watchlist. clear( ); } 

   private:
      size_t firstconsistent( 
         const std::function< bool( const lemma& , size_t ) > & consistency,
         const lemma& lem );

      void checkwatching( variable v, size_t& current_confl,
         const std::function< bool( const lemma& , size_t ) > & consistency );
       
      void addwatch( size_t lem, size_t i );
      void removewatch( size_t lem, size_t i );
         // Start/stop watching lem/i.

      static void insert( std::vector< watched > & watch, watched w );
      static void remove( std::vector< watched > & watch, watched w );
         // We assume that there is only one occurrence. 
   };


   inline std::ostream& 
   operator << ( std::ostream& out, lemmasystem::watched w )
   {
      out << w.lem << "/" << w.i;
      return out;
   }

   std::ostream& 
   operator << ( std::ostream& out, 
                 const std::vector< lemmasystem::watched > & watchlist );

   std::ostream& operator << ( std::ostream& out, const lemmasystem& sys );

}}

#endif


