
// Part of Geo III, written by Hans de Nivelle, June 2016.

#include "refinement.h"
#include "resolution.h"

namespace geometric 
{
namespace constraints
{


   std::ostream& operator << ( std::ostream& out, const refinement& ref )
   {
      out << "Using additional parents {";
      for( auto p = ref. parents. begin( ); p != ref. parents. end( ); ++ p )
      {
         if( p != ref. parents. begin( ))
            out << ", ";
         else
            out << " ";
         out << *p;
      }
      out << " },\n"; 

      out << "and the substitution having size " << ref. subst_restore << ",\n";
      out << "clause " << ref. cl << " was refined from " << ref. previous;
      out << " into " << ref. current << "\n";

      return out;
   }


   std::ostream& operator << ( std::ostream& out, 
                               const std::vector< refinement > & log )
   {
      out << "Refinement Stack:\n";
      for( size_t i = 0; i < log. size( ); ++ i )
      {
         out << i << " :\n";
         out << log[i] << "\n";
      }
      return out;
   }


   bool refine( subststack< size_t > & theta, gcsp& g, 
                size_t cl, size_t cl1,
                std::vector< refinement > & log )
   {
      auto initial = g. clauses[ cl ]. current;

      size_t i = g. clauses[ cl ]. current. s0;
      while( i < g. clauses[ cl ]. current. s1 )
      {
         if( solvable( theta, g, g. clauses[ cl ], i, g. clauses[ cl1 ] ))
            ++ i;
         else 
         {
            g. clauses[ cl ]. disable(i);
         }
      }

      if( initial. size( ) != g. clauses[cl]. current. size( ))
      {
         // Something was changed, so we write the change into log:

         log. push_back( { cl, { cl1 }, 
                           initial, g. clauses[ cl ]. current, 
                           theta. size( ) } );
      }

      return unit_empty( g. clauses[cl], theta );
   }

   
   bool refine( subststack< size_t > & theta, gcsp& g,
                size_t cl, size_t cl1, size_t cl2, 
                std::vector< refinement > & log )
   {
      auto initial = g. clauses[ cl ]. current;

      size_t i = g. clauses[ cl ]. current. s0;
      while( i < g. clauses[ cl ]. current. s1 )
      {
         if( solvable( theta, g, g. clauses[ cl ], i, 
                                 g. clauses[ cl1 ], g. clauses[ cl2 ] ))
         {
            ++ i;
         }
         else
         {
            g. clauses[ cl ]. disable(i);
         }
      }

      if( initial. size( ) != g. clauses[cl]. current. size( ))
      {
         // Something was changed, so we write the change into log:

         log. push_back( { cl, { cl1, cl2 }, 
                           initial, g. clauses[ cl ]. current,
                           theta. size( ) } );
      }

      return unit_empty( g. clauses[cl], theta );
   }


   namespace 
   {
      bool sharevariable( const gcsp::clause& cl1, const gcsp::clause& cl2 )
      {
         for( size_t v1 = 0; v1 < cl1. nrvars( ); ++ v1 )
         for( size_t v2 = 0; v2 < cl2. nrvars( ); ++ v2 )
         {
            if( cl1. var( v1 ) == cl2. var( v2 ))
               return true;
         }
         return false;
      }
   }
 
 
   bool refine( subststack< size_t > & theta, gcsp& g, 
                size_t local, 
                std::vector< refinement > & log,
                size_t k )
   {

      if( local < 2 ) 
         return true;

      // We check subsets of size 2:

      {
         // Note that log may grow in the process. 

         for( size_t pos = k; pos < log. size( ); ++ pos )
         {

            if( log[ pos ]. iscurrent(g))
            {
               size_t cl = log[ pos ]. cl;

               for( size_t v = 0; v < g. clauses[ cl ]. nrvars( ); ++ v )
               {
                  const auto& occ = g. inclauses( g. clauses[cl]. var(v) );
                  for( auto cl1 : occ )
                  {
                     if( cl1 != cl )
                     {
                        auto b = refine( theta, g, cl1, cl, log );
                        if( !b ) return false;
                     } 
                  }
               }
            }
         }
      }

      if( local < 3 )
         return true;

      // We check subsets of size 3:

      {
         // Note that log may grow:

         for( size_t pos = k; pos < log. size( ); ++ pos )
         {
            if( log[ pos ]. iscurrent(g)) 
            {
               size_t cl = log[ pos ]. cl;

               for( size_t v1 = 0; v1 < g. clauses[ cl ]. nrvars( ); ++ v1 )
               for( size_t v2 = 0; v2 < g. clauses[ cl ]. nrvars( ); ++ v2 )
               {
                  const auto& occ1 = g. inclauses( g. clauses[cl]. var(v1) );
                  const auto& occ2 = g. inclauses( g. clauses[cl]. var(v2) );
                  {
                     for( auto cl1: occ1 )
                     for( auto cl2: occ2 )
                     {
                        if( cl != cl1 && cl != cl2 && cl1 < cl2 && 
                            sharevariable( g. clauses[cl1], g. clauses[cl2] ))
                        {
                           // std::cout << "[ " << cl << ", " << cl1 << ", ";
                           // std::cout << cl2 << " ]\n";

                           auto b = refine( theta, g, cl1, cl, cl2, log );
                           if( !b ) return false;

                           b = refine( theta, g, cl2, cl, cl1, log );
                           if( !b ) return false; 
                        }
                     }
                  }
               }
            }
         }
      }

      // 4-cycles can also be considered. More probably makes no sense.

      return true;
   }


   std::vector< refinement > 
   allclauses( const gcsp& g, const subststack< size_t > & theta )
   {
      std::vector< refinement > log;

      for( size_t cl = 0; cl < g. clauses. size( ); ++ cl )
      {
         log. push_back( { cl, {}, 
                           g. clauses[ cl ]. current,
                           g. clauses[ cl ]. current, theta. size( ) } );   
      }

      return log;
   }

}}


