
// Written by Hans de Nivelle, May 2016.

// A refinement is when some clause is replaced by a subclause.
// There are two possible ways in which this can happen:
// By guessing or by filtering.

#ifndef GEOMETRIC_CONSTRAINTS_REFINEMENT_INCLUDED
#define GEOMETRIC_CONSTRAINTS_REFINEMENT_INCLUDED  1

#include <list>
#include "gcsp.h"
#include "lemma.h"


namespace geometric
{
namespace constraints 
{

    
   struct refinement 
   {
      // Every refinement has a main clause that is being refined.
      // This applies to decisions and to refinements obtained by filtering.
      // Refinements obtained by filtering can be recognized from
      // the fact that they have at least one parent. 
 
      size_t cl;                         // Clause that is being refined.
      std::list< size_t > parents;       // Side parents. No parents=decision.

      gcsp::clause::state previous;     
      gcsp::clause::state current;

      size_t subst_restore;   

      std::list< lemma > conflict_lemmas;
                           // The conflict lemmas that we managed to 
                           // collect so far. Only lemmas that are not
                           // in the lemmasystem need to be stored. 


      refinement( size_t cl, 
                  const std::list< size_t > & parents,
                  gcsp::clause::state previous,
                  gcsp::clause::state current,
                  size_t subst_restore )
         : cl{ cl },
           parents{ parents },
           previous{ previous },
           current{ current },
           subst_restore{ subst_restore } 
      { }
     
      bool iscurrent( const gcsp& g ) const  
         { return current. size( ) == g. clauses[ cl ]. current. size( ); }

   };


   std::ostream& operator << ( std::ostream& , const refinement& ref );
   std::ostream& 
   operator << ( std::ostream& , const std::vector< refinement > & log ); 


   // Check if cl is unit or empty. If cl is unit, extend theta. 
   // If cl is empty, return false.

   inline 
   bool unit_empty( const gcsp::clause& cl, subststack< size_t > & theta )
   {
      if( cl. current. size( ) == 1 )
         match( theta, cl, cl. current. s0 );
      return cl. current. size( ) != 0;
   }
          
   bool refine( subststack< size_t > & theta, gcsp& g,
                size_t cl, size_t cl1,
                std::vector< refinement > & log );

      // Refine clause cl against cl1, and write the refinement in log. 
      // Return false if cl becomes empty, and extend theta if cl becomes
      // unit. 

   bool refine( subststack< size_t > & theta, gcsp& g,
                size_t cl, size_t cl1, size_t cl2,
                std::vector< refinement > & log );

      // Refine clause cl against cl1,cl2, and write the refinement in log. 
      // Return false if cl becomes empty, and extend theta if cl becomes
      // unit. 

   bool refine( subststack< size_t > & theta, gcsp& g, 
                size_t local, 
                std::vector< refinement > & log, size_t k );

      // Refine clauses in g as much as possible, considering connected 
      // subsets with size determined by local, which have to 
      // involving a clause in log [ k .. log. end( ) ]. 
      // In case a clause becomes unit, s is extended accordingly.
      // In case a clause becomes empty, we return false.

   bool refine( gcsp& g, size_t cl, size_t cl1,
                std::vector< refinement > & log );

      // Refine clause cl against cl1, and write the refinement in log.
      // Return false if cl becomes empty, and extend theta if cl becomes
      // unit.

   bool refine( gcsp& g, size_t cl, size_t cl1, size_t cl2,
                std::vector< refinement > & log );

      // Refine clause cl against cl1,cl2, and write the refinement in log.
      // Return false if cl becomes empty, and extend theta if cl becomes
      // unit.

   bool refine( gcsp& g, size_t local,
                std::vector< refinement > & log, size_t k );

      // Refine clauses in g as much as possible, considering connected
      // subsets with size determined by local, which have to
      // involving a clause in log [ k .. log. end( ) ].
      // In case a clause becomes unit, s is extended accordingly.
      // In case a clause becomes empty, we return false.












   std::vector< refinement > 
   allclauses( const gcsp& g, const subststack< size_t > & theta );
      // Create a stack that states that nothing happened on any
      // of the clauses. 

}}

#endif

