
// Written by Hans de Nivelle, June 2016.

#include "solve.h"
#include "resolution.h"


namespace geometric
{
namespace constraints 
{

   namespace
   {

      static constexpr bool checksolutions = true;
         // Checking correctness of solutions can be costly when there are
         // many.

      static constexpr unsigned int outputlevel = 1;
         // Higher number means more output.
         // 0 : print nothing.
         // 1 : print one in 10000 solutions.
         // 2 : print all solutions.
         // 3 : print even more. 

      static constexpr bool stopfirst = true;
         // Makes search algorithms stop on first solution.
         // Otherwise, all solutions get enumerated. 
         // Normal operation should stop on the first solution. 
   }


   void printsolution( const gcsp& g, const subststack< size_t > & theta,
                       size_t& nrsolutions )
   {
      if( outputlevel )
      {
         if( outputlevel > 1 || nrsolutions % 10000 == 0 )
         {
            std::cout << "Solution " << nrsolutions << ":\n";
            std::cout << theta << "\n\n";
         }
      }

      if( checksolutions && !g. issolution( theta, true ))
      {
         std::cout << "-------------------------\n";
         std::cout << g << "\n";
         std::cout << theta << "\n";

         std::cout << "substitution is not a solution!\n\n";
         throw std::runtime_error( "found incorrect solution!\n" );
      }

      ++ nrsolutions;
   }


   size_t pick( const gcsp& g )
   {
      size_t best = g. clauses. size( );
      size_t bestsize = 0;

      for( size_t cl = 0; cl < g. clauses. size( ); ++ cl )
      {
         size_t s = g. clauses[ cl ]. current. size( );
         if( s > 1 &&
             ( best == g. clauses. size( ) || s < bestsize ))
         {
            best = cl;
            bestsize = s;
         }
      }
      return best;
   }


   void naive::solve( const gcsp& g,
                      subststack< size_t > & theta, 
                      size_t& nrsolutions,
                      size_t level )
   {
      if( level == g. clauses. size( ))
      {
         printsolution( g, theta, nrsolutions ); 
      } 
      else
      {
         size_t s = theta. size( );

         for( size_t j = g. clauses[ level ]. current. s0; 
                     j < g. clauses[ level ]. current. s1; ++ j )
         {
            if( match( theta, g. clauses[ level ], j ) && 
                !g. impliesblocked( theta, s ))
            {
               solve( g, theta, nrsolutions, level + 1 );
            }

            if( nrsolutions && stopfirst )
               return;

            theta. restore(s);
         }
      }
   }


   namespace
   {
      // restore_confl should of course be merged with derive. 
      // In the main clause, it is possible that substlets were switched
      // off, because of indirect conflicts through a lemma or blocking.
      // These substlets have to be resolved away with the lemma (or blocking).
      // Currently, it is implemented as post processing, but it should
      // be done before resolution on the main clause. 

      lemma 
      restore_confl( const lemma& lem,
                     subststack< size_t > & theta, 
                     lemmasystem& sys, const gcsp& g )
      {
         lemma res;
 
         size_t s = theta. size( );

         for( size_t i = 0; i < lem. nrsubstlets( ); ++ i )
         {
            if( !match( theta, lem, i ))
            {
               res. insert( lem, i );
               goto found;
            }

            {
               size_t l = sys. findconflict( theta, s );
               if( l != sys. size( ))
               {
                  for( size_t j = 0; j < sys[l]. nrsubstlets( ); ++ j )
                  {
                     if( consistent( lem, i, sys[l], j ))
                        res. insert( sys[l], j );
                  }
                  goto found; 
               }
            }

            {
               auto bl = g. getblocked( theta, s );
               if( bl. first != g. blockings. size( ))
               {
                  lemma sigma;
                  add_sigma_resolvent( sigma, g,
                                       g. blockings[ bl. first ], bl. second );

                  for( size_t j = 0; j < sigma. nrsubstlets( ); ++ j )
                  {
                     if( consistent( lem, i, sigma, j ))
                        res. insert( sigma, j );
                  }
                  goto found;
               }
            }

            std::cout << lem << "/" << i << "\n";
            throw std::runtime_error( "restore_confl could not restore" ); 
         found:
            theta. restore(s); 
         }
         return res;
      }
   }



   // We assume that theta does not imply a blocking or lemma when called,
   // and that g does not contain an empty clause. 

   std::list<conflict_lemma>
   learning::forward( gcsp& g, lemmasystem& sys, 
                      subststack< size_t > & theta, size_t s,
                      std::vector< refinement > & stack )
   {
      while( s < theta. size( ))
      {
         variable v = theta[s]. first;

         const auto& occ = g. inclauses(v);
         for( auto c : occ )
         {
            // It is not a for loop, because we either delete or 
            // move to the next.

            auto original_state = g. clauses[c]. current;

            size_t i = g. clauses[c]. current. s0;  
            while( i != g. clauses[c]. current. s1 ) 
            {
               if( consistent( theta, g. clauses[c], i ) && 
                   ! g.impliesblocked( theta, g. clauses[c], i ))
               {

                  // Because method findconflict( theta, g.clauses[c], i ) 
                  // is missing in lemmasystem, we have to match. 

                  size_t ss = theta. size( );
                  match( theta, g. clauses[c], i );
                  if( sys. findconflict( theta, ss ) == sys. size( )) 
                     ++ i;
                  else
                  {
                     g. clauses[c]. disable(i); 
                  }
                  theta. restore(ss); 
               }
               else
               {
                  g. clauses[c]. disable(i); 
               }
            }

            if( original_state. size( ) != g. clauses[c]. current. size( ))
            {
               if( g. clauses[c]. current. size( ) == 0 )
               {
                  g. clauses[c]. current = original_state; 
                  return { conflict_lemma( 
                       restore_confl( g. clauses[c], theta, sys, g )) };
               }

               // Note that c is not the real parent. 

               stack. push_back( { c, {c}, original_state, 
                                   g. clauses[c]. current, theta. size( ) } ); 

               if( g. clauses[c]. current. size( ) == 1 )
               {
                  size_t sss = theta. size( );
                  match( theta, g.clauses[c], g. clauses[c]. current. s0 );

                  // I think that it cannot happen, but we have to check
                  // lemmas, because sys may be watching variables in the 
                  // extension of theta. If we don't correct the watchings,
                  // we loose them.  

                  if( sys. findconflict( theta, sss ) != sys. size( ))
                  {
                     throw std::runtime_error( "this surprises me" );
                  }
               }
            }
         }
         ++ s; 
      }

      return { };
   }

   namespace
   {

      bool 
      conflicts( subststack< size_t > & theta, const lemma& lem )
      {
         for( size_t i = 0; i < lem. nrsubstlets( ); ++ i )
         {
            if( consistent( theta, lem, i ))
               return false;
         }
         return true;
      }

    
      // Find the shortest conflict lemma of theta among lemmas:

      std::list< lemma > :: const_iterator
      findconflict( subststack< size_t > & theta, 
                    const std::list< lemma > & lemmas )
      {
         std::list< lemma > :: const_iterator res = lemmas. end( );
         for( auto p = lemmas. begin( ); p != lemmas. end( ); ++ p )
         {
            if( res != lemmas. end( ))
            {
               if( res -> nrsubstlets( ) > p -> nrsubstlets( ))
               {
                  if( conflicts( theta, *p ))
                     res = p;
               }
            }
            else
            {
               if( conflicts( theta, *p ))
                  res = p;
            }
         }
         return res;
      }


      std::list<conflict_lemma> 
      derive( gcsp& g, lemmasystem& sys, 
              subststack< size_t > & theta,
              const gcsp::clause& cl, 
              const std::list< lemma > & lemmas )
      {

         // In the main clause, every atom that is turned off should
         // conflict theta, or imply a lemma or blocking.

         // if( sys. findconflict( theta, 0 ) < sys. size( ))
         //    throw std::runtime_error( "implies lemma alone" );

         // This must be made into a resolution rule. Now it is only a check.

         for( size_t i = cl. initial. s0;
                     i != cl. initial. s1;
                     ++ i )
         {
            if( !cl. current. contains(i))
            {
               size_t s = theta. size( );
               if( match( theta, cl, i ))
               {
                  if( !g.impliesblocked( theta, s ) &&
                      sys. findconflict( theta, s ) == sys. size( ))
                  {
                     std::cout << cl << "\n";
                     std::cout << "Substlet " << i;
                     std::cout << " was wrongly switched off\n";
                     throw std::runtime_error( 
                                "atom incorrectly switched off" ); 
                  }
               }
               theta. restore(s);
            }
         }

         // We are going to collect the conflict lemmas:

         std::vector< const lemma* > lemma_pointers;   // Needed by resolve.
         std::list< lemma > sigmas;         
 
         size_t s = theta. size( ); 
         for( size_t i = cl.current.s0; i != cl.current.s1; ++ i )
         {
            if( !match( theta, cl, i ))
               throw std::runtime_error( "should be impossible" );

            const lemma* pointer = nullptr;
            {
               size_t l = sys. findconflict( theta, s );
               if( l != sys. size( ))
               {
                  pointer = & sys[l];
                  goto conflict_found; 
               }
            }

            { 
               auto bl = g. getblocked( theta, s );  
               if( bl. first != g. blockings. size( ))
               {
                  sigmas. push_back( {} );
                  add_sigma_resolvent( sigmas. back( ), g, 
                                       g. blockings[ bl. first ], bl. second );
                  if( !conflicts( theta, sigmas. back( )))
                     throw std::runtime_error( "sigma-resolution failed" );
         
                  pointer = & sigmas. back( ); 
                  goto conflict_found; 
               }
            }

            {
               auto p = findconflict( theta, lemmas );
               if( p == lemmas. end( ))
               {
                  std::cout << theta << "\n";
                  std::cout << cl << "/" << i << "\n";
                  throw std::runtime_error( "derive: Failed to find conflict" );
               }
               else
               {
                  pointer = & (*p);
                  goto conflict_found;
               }
            }

            throw std::runtime_error( "Failed to find a conflict lemma" );

         conflict_found: 
            lemma_pointers. push_back( pointer );  
            theta. restore(s);
         }

         return { resolvent( &cl, lemma_pointers, cl. current, { } ) };
      }
   }
     

   // We assume that forward reasoning has been exhaustively applied.
   // If stopfirst == true, we stop on the first solution, and this solution
   // remains in theta.

   std::list<conflict_lemma>
   learning::solve( gcsp& g, lemmasystem& sys, 
                    subststack< size_t > & theta, 
                    std::vector< refinement > & stack, 
                    size_t& nrsolutions )
   {
      size_t picked = pick(g); 

      if( picked == g. clauses. size( ))
      {
         ++ nrsolutions; 
         return { }; 
         printsolution( g, theta, nrsolutions ); 
         return { };
      }
      else
      {
         size_t nrsolutions2 = nrsolutions; 
            // So that we know if there were solutions.

         auto restore = g. clauses[ picked ]. current;

         size_t s = theta. size( );
         size_t k = stack. size( ); 

         stack. push_back( { picked, {}, restore, restore, s } );

         for( stack. back( ). current = restore. first(1); 
              stack. back( ). current. size( );
              stack. back( ). current. next( restore )) 
         {
            if( !match( theta, g. clauses[ picked ], 
                        stack. back( ). current. s0 ))
            {
               std::cout << "theta " << theta << "\n";
               throw std::runtime_error( "could not match choice" );
            }

            if( !g. impliesblocked( theta, s ) &&
                sys. findconflict( theta, s ) == sys. size( ))
            {
               g. clauses[ picked ]. current = stack. back( ). current; 
               
               auto confl = forward( g, sys, theta, s, stack );
               if( confl. size( ))
               {
                  if( !conflicts( theta, confl. front( ). getlemma( sys )))
                     throw std::runtime_error( "not a conflict lemma" );
               }
 
               if( confl. size( ) == 0 )
               { 
                  confl = solve( g, sys, theta, stack, nrsolutions ); 
                  if( nrsolutions && stopfirst )
                     return { };
               }

               if( confl. size( ))
               {
                  while( stack. size( ) > k+1 )
                  {
                     if( stack. back( ). parents. size( ) != 1 )
                        throw std::runtime_error( "expected deterministic" );
    
                     theta. restore( stack. back( ). subst_restore );

                     if( !conflicts( theta, confl. front( ). getlemma( sys )))
                     { 
                        if( confl. front( ). wascreated( ))
                        {
                           size_t s = 
                              sys. insert( confl. front( ). getlemma( sys ),
                                        theta );
     
                           confl. front( ) = conflict_lemma(s);  
                        }

                        std::list< lemma > parent;
                        if( confl. front( ). wascreated( ))
                        {
                           parent. push_back( std::move( confl. front( )). 
                                              getlemma( sys )); 
                        }
                        confl = derive( g, sys, theta, 
                                   g. clauses[ stack. back( ). cl ], 
                                   parent ); 

                        
                        if( !conflicts( theta, 
                                        confl. front( ).  getlemma( sys )))
                        {
                           confl = { restore_confl( 
                                        confl. front( ). getlemma(sys), 
                                        theta, sys, g ) };
                        }

                        if( !conflicts( theta,
                                        confl. front( ).  getlemma( sys )))
                        {
   throw std::runtime_error( "deterministic resolvent not a conflict lemma" );
                        }

                     }

                     g. clauses[ stack. back( ). cl ]. current  
                              = stack. back( ). previous;
                     stack. pop_back( );
                  }

                  if( stack. size( ) != k+1 || stack. back( ). parents. size( ))
                     throw std::runtime_error( "problems in stack" );  

                  theta. restore( stack. back( ). subst_restore );
                  if( conflicts( theta, confl. front( ). getlemma( sys )))
                  {
                     g. clauses[ picked ]. current = restore;    
                     stack. pop_back( ); 
                     return std::move( confl ); 
                  }

                  if( confl. front( ). wascreated( )) 
                  {
                     stack. back( ). conflict_lemmas. 
                        push_back( std::move( confl. front()). getlemma(sys)); 
                  }
               }
               else
               {
                  if( nrsolutions == nrsolutions2 )
                     throw std::runtime_error( "no lemma but no solutions" );

                  while( stack. size( ) > k+1 )
                  {
                     g. clauses[ stack. back( ). cl ]. current =
                          stack. back( ). previous; 
                     stack. pop_back( ); 
                  }

                  theta. restore( stack. back( ). subst_restore );
               }
            }
            theta. restore( stack. back( ). subst_restore ); 
         }

         g. clauses[ picked ]. current = restore;

         if( nrsolutions != nrsolutions2 )
         {
            stack. pop_back( ); 
            return { };
         } 

         auto confl = derive( g, sys, theta,
                              g. clauses[ picked ],
                              stack. back( ). conflict_lemmas );

         confl = 
         { restore_confl( confl. front( ). getlemma(sys), theta, sys, g ) }; 

         if( !conflicts( theta, confl. front( ). getlemma( sys )))
         {
            std::cout << confl. front( ). getlemma( sys ) << "\n";
            for( size_t i = 0; 
                 i != confl. front( ). getlemma( sys ). nrsubstlets( );
                 ++ i )
            {
               if( consistent( theta, confl. front( ). getlemma(sys), i ))
                  std::cout << i << " is consistent\n";
            }
            throw std::runtime_error( "not a conflict lemma after choices" );
         }

         stack. pop_back( ); 
         return { std::move( confl ) }; 
      }
   }




   // We assume that theta does not imply a blocking when called,
   // and that g does not contain an empty clause. 

   bool notlearning::forward( gcsp& g, 
                              subststack< size_t > & theta, size_t s,
                              std::vector< refinement > & stack )
   {
      while( s < theta. size( ))
      {
         variable v = theta[s]. first;

         const auto& occ = g. inclauses(v);
         for( auto c : occ )
         {
            // It is not a for loop, because we either delete or 
            // move to the next.

            auto original_state = g. clauses[c]. current;

            size_t i = g. clauses[c]. current. s0;  
            while( i != g. clauses[c]. current. s1 ) 
            {
               if( consistent( theta, g. clauses[c], i ) &&
                   ! g.impliesblocked( theta, g. clauses[c], i ))
               {
                  ++ i;
               }
               else
               {
                  g. clauses[c]. disable(i); 
               }
            }

            if( original_state. size( ) != g. clauses[c]. current. size( ))
            {
               if( g. clauses[c]. current. size( ) == 0 )
               {
                  g. clauses[c]. current = original_state; 
                  return false;
               } 

               stack. push_back( { c, {c}, original_state, 
                                   g. clauses[c]. current, theta. size( ) } ); 

               if( g. clauses[c]. current. size( ) == 1 )
               {
                  match( theta, g.clauses[c], g. clauses[c]. current. s0 );
               }
            }
         }
         ++ s; 
      }
      return true;
   }


   // We assume that forward reasoning has been exhaustively applied.
   // If stopfirst == true, we stop on the first solution, and this solution
   // remains in theta.

   void notlearning::solve( gcsp& g, 
                            subststack< size_t > & theta, 
                            std::vector< refinement > & stack, 
                            size_t& nrsolutions )
   {
      size_t picked = pick(g); 

      if( picked == g. clauses. size( ))
      {
         ++ nrsolutions;
         return; 
         printsolution( g, theta, nrsolutions ); 
      }
      else
      {
         auto restore = g. clauses[ picked ]. current;
         for( auto choice = restore. first(1); 
                   choice. size( );
                   choice. next( restore )) 
         {
            size_t s = theta. size( );
            size_t k = stack. size( ); 

            if( !match( theta, g. clauses[ picked ], choice. s0 ))
            {
               std::cout << "theta " << theta << "\n";
               throw std::runtime_error( "choice was unmatchable!" );
            }

            if( !g. impliesblocked( theta, s ))
            {
               g. clauses[ picked ]. current = choice; 

               stack. push_back( { picked, {}, restore, choice, s } );

               if( forward( g, theta, s, stack ))
                  solve( g, theta, stack, nrsolutions ); 
           
               if( stopfirst && nrsolutions )
                  return;
 
               while( stack. size( ) > k )
               {
                  g. clauses[ stack. back( ). cl ]. current
                         = stack. back( ). previous; 
                  stack. pop_back( ); 
               }
            }
            theta. restore(s);
         }
      }
   }



   std::list< subststack< size_t >> bestsolution( gcsp& g )
   {

      g. initfromalpha( nullptr );

      subststack< size_t > theta;

      if( !preprocessor( g, theta, 2 )) 
         return { }; 

      g. hard_commit( );  
      g. set_initials( );
 
      size_t s = theta. size( );
         // At this point, we know that theta must be included in 
         // every solution. In the future, we will restore theta to s. 

      size_t nrsolutions = 0; 
      std::vector< refinement > stack;
      lemmasystem sys;

      learning::solve( g, sys, theta, stack, nrsolutions ); 

      if( nrsolutions )
      {
         if( checksolutions && !g.issolution( theta, true ))
            throw std::runtime_error( "wrong solution" );

         auto best_theta = theta; 
         set< size_t > alpha = g. getassumptions( best_theta );

         // std::cout << "the first solution has: " << alpha << "\n";
         // std::cout << "best theta = " << best_theta << "\n";
 
         if( alpha. size( ) == 0 )
            return { best_theta };   // We won't find better. 

         size_t k = alpha. max( ) + 1;

         while(k)
         {
            -- k; 

            if( alpha[k] )
            {
               // std::cout << "considering " << k << "\n";
               // std::cout << "alpha = " << alpha << "\n";

               auto alpha2 = alpha;
               alpha2. erase(k);
               for( size_t j = 0; j < k; ++ j )
                  alpha2. insert(j);

               g. initfromalpha( &alpha2 );
               theta. restore(s);

               if( preprocessor( g, theta, 2 ))
               {
                  // std::cout << "Made it through the first checks:\n";
                  // std::cout << cs << "\n";
                  // std::cout << theta << "\n";

                  nrsolutions = 0;
                  stack. clear( ); 
                  sys. clear( );
                  g. set_initials( ); 
                  learning::solve( g, sys, theta, stack, nrsolutions );

                  if( nrsolutions ) 
                  {
                     if( checksolutions && !g. issolution( theta, true ))
                        throw std::runtime_error( "wrong solution" );

                     best_theta = theta;
                     alpha = g. getassumptions( best_theta );
                  }
               }
            }
         }

         return { best_theta }; 
      }
      else
         return { }; 
   } 


}}


