
// Written by Hans de Nivelle, Jan-June 2016.
// A few solvers, both learning and non-learning, and squeezing. 

#ifndef GEOMETRIC_CONSTRAINTS_SOLVE_INCLUDED
#define GEOMETRIC_CONSTRAINTS_SOLVE_INCLUDED  1

#include "gcsp.h"
#include "refinement.h"
#include "preprocessor.h"
#include "lemmasystem.h"


namespace geometric
{
namespace constraints 
{

   size_t pick( const gcsp& g );
      // Pick a shortest clause from the clauses that have more
      // than one choice. 
      // Return g. clauses. size( ) if all clauses have length
      // 0 or 1. 

   // All solving functions assume that g has been initialized, and 
   // preprocessed. 

   void printsolution( const gcsp& g, const subststack< size_t > & theta,
                       size_t& nrsolutions ); 



   // A conflict lemma is either new (created), or it originates from
   // a lemma system. 

   struct conflict_lemma
   {
      std::list< lemma > created;
      size_t index;

      conflict_lemma( const lemma& lem )
         : created{ { lem } }
      { }

      conflict_lemma( lemma&& lem )
      {
         created. push_back( std::move(lem) );
      }

      conflict_lemma( size_t index )
         : index{ index }
      { }

      bool wascreated( ) const { return created. size( ); }
         // Rename into notinsystem?

      const lemma& getlemma( const lemmasystem& sys ) const &
         { return created. size( ) ? created. front( ) : sys[ index ]; }
            // Get the lemma.

      lemma getlemma( const lemmasystem& sys ) &&
         { return created. size( ) ?
              std::move( created. front( )) : sys [ index ]; }

   };



   namespace naive
   { 
      void solve( const gcsp& g, subststack< size_t > & theta,
                  size_t& nrsolutions, size_t level ); 

      inline void solve( gcsp& g, subststack< size_t > & theta )
      {
         if( !preprocessor( g, theta, 2 ))
            return;

         size_t nrsolutions = 0;
         solve( g, theta, nrsolutions, 0 );
      }
                  
   }


   namespace notlearning
   {

      bool forward( gcsp& g,
                    subststack< size_t > & theta, size_t s,
                    std::vector< refinement > & stack );

      void solve( gcsp& g, subststack< size_t > & theta,
                  std::vector< refinement > & stack, size_t& nrsolutions );

      inline void solve( gcsp& g, subststack< size_t > & theta )
      {
         if( !preprocessor( g, theta, 2 ))
         {
            return;
         }

         size_t nrsolutions = 0;
         std::vector< refinement > stack;
         solve( g, theta, stack, nrsolutions );
      }
   }


   namespace learning
   {

      std::list<conflict_lemma>
      forward( gcsp& g, lemmasystem& sys, 
               subststack< size_t > & theta, size_t s,
               std::vector< refinement > & stack );

      std::list<conflict_lemma>
      solve( gcsp& g, lemmasystem& sys,
             subststack< size_t > & theta, 
             std::vector< refinement > & stack, size_t& nrsolutions );

      inline void solve( gcsp& g, 
                         subststack< size_t > & theta )
      {
         if( !preprocessor( g, theta, 2 ))
         {
            return;
         }

         g. set_initials( );

         size_t nrsolutions = 0;
         std::vector< refinement > stack;
         lemmasystem sys;
         auto lem = solve( g, sys, theta, stack, nrsolutions );
      }

   }


   std::list< subststack< size_t >> bestsolution( gcsp& g );    
      // List contains 0 or 1 elements.
      // g cannot be const, because we sometimes reset the usable  
      // choices in g. 

}}

#endif

