
// There is always the dilemma between long and short substitutions.
// Long substitutions should use an index for quick lookup, but it may take 
// time to build this index, and is relatively costly to make changes to
// the index. 

// Short substitutions should use simple linear lookup. 
// An empty linear lookup can be easily created. Adding and removing
// is also very cheap.
// 
// We try to solve this dilemma by using a combination in which a small
// part of the substitution is kept on a stack, and the rest is kept
// in an index.


#ifndef GEOMETRIC_CONSTRAINTS_SUBSTSTACK_DEFINED
#define GEOMETRIC_CONSTRAINTS_SUBSTSTACK_DEFINED   1

#include "../variable.h"
#include <unordered_map>
#include <iostream>
#include <functional>


namespace geometric 
{ 
namespace constraints 
{

   template< typename D >
   class subststack
   {
      
      std::unordered_map< variable, D,
                          variable::hash, variable::equal_to > index;
         // An index to the occurrences of a variable in
         // clauses and blockings.
 
      size_t inindex;
         // Stack entries between 0 and inindex occur also in index.
          
      std::vector< std::pair< variable, D >> stack;

      static constexpr size_t maxdiff = 5;
         // Maximum number of unindexed entries.

   public:
      subststack( ) 
         : inindex{0}
      { }

      void clear( )
      {
         index. clear( );
         inindex = 0;
         stack. clear( );
      }
     
      // Don't overwrite an existing value. That won't work well.
 
      void assign( variable v, const D& d )
      {
         stack. push_back( { v, d } );
         while( stack. size( ) > index. size( ) + maxdiff )
         {
            index. insert( stack[ inindex ++ ] );
         }
      }

      const std::pair< variable, D > & operator [] ( size_t i ) const
         { return stack[i]; } 
      
      void print( std::ostream& out ) const;

      const D* lookup( variable v ) const;
      
      size_t size( ) const { return stack. size( ); }
         
      void restore( size_t k )
      {
         while( stack. size( ) > k ) 
         {
            if( stack. size( ) == inindex )
            {
               index. erase( stack. back( ). first );
               -- inindex;
            }
            stack. pop_back( );
         }
      }

      template< typename D1 > 
      subststack< D1 > 
      convertrange( const std::function< D1(D) > & conv ) const;
         // Convert the range to type D1. 
         //    For every assignment v := d, the result has v := conv(d). 
   };


   template< typename D >
   const D* subststack<D> :: lookup( variable v ) const
   {
      // First look in index:

      auto p = index. find(v);
      if( p != index. end( ))
         return & ( p -> second );

      // After that, in the stack (the part that is not in index):

      for( auto p = stack. begin( ) + inindex; p != stack. end( ); ++ p )
      {
         if( p -> first == v ) 
            return & ( p -> second );
      }

      return nullptr;
   }


   template< typename D >
   void subststack<D> :: print( std::ostream& out ) const
   {
      out << "subststack(";
      for( auto p = stack. begin( ); p != stack. end( ); ++ p )
      {
         if( p != stack. begin( ))
            out << ", ";
         else
            out << " ";
         out << ( p -> first ) << " := " << ( p -> second );
      }
      out << " )";

#if 0
      out << "\ninindex: " << inindex << "\n";

      out << "\nHashmap:\n";
      for( const auto& p : index )
         out << "   " << p. first << " --> " << p. second << "\n";
#endif
   }

 
   template< typename D > 
   template< typename D1 > 
   subststack< D1 > 
   subststack< D > :: convertrange( 
             const std::function< D1(D) > & conv ) const
   {
      subststack<D1> res; 
      for( auto p = stack. cbegin( ); p != stack. cend( ); ++ p )
         res. assign( p -> first, conv( p -> second ));

      return res; 
   }


   template< typename D >
   std::ostream& operator << ( std::ostream& out, const subststack<D> & s )
   {
      s. print( out );
      return out;
   }

}}


#endif


