
#include "parser.h"

namespace geometric {
namespace parser {

using syntaxerror = ::parser::syntaxerror;
using conversionerror = ::parser::conversionerror;


variable makevariable( const std::vector< std::string > & ident,
                       const ::parser::inputstream& input ) 
{
   if( ident. size( ) != 1 )
   {
      throw syntaxerror( tostring( "unable to convert compound identifier ",
                                   ident[0], " :: ... into variable" ), 
                         input );
   }

   // We accept from V0 to V9999:

   if( ident[0]. size( ) < 2 || ident[0]. size( ) > 5 ||
       ident[0][0] != 'V' )
   {
      throw syntaxerror( 
              tostring( "unable to convert identifier ",
                        ident[0],
                        " into variable" ), input );
   }

   long unsigned int k = 
      make_unsigned( ident[0]. begin( ) + 1, ident[0]. end( ), input );

   variable v;
   while( k )
   { 
        ++ v;
        -- k;
   }
   return v; 
}


element makeelement( const std::vector< std::string > & ident,
                     const ::parser::inputstream& input )
{
   if( ident. size( ) != 1 )
   {
      throw syntaxerror( tostring( "unable to convert compound identifier ",
                                   ident[0], " :: ... into element" ),
                         input );
   }

   // We accept from E0 to E9999:

   if( ident[0]. size( ) < 2 || ident[0]. size( ) > 5 ||
       ident[0][0] != 'E' )
   {
      throw syntaxerror(
              tostring( "unable to convert identifier ",
                        ident[0],
                        " into element" ), input );
   }

   long unsigned int k =
      make_unsigned( ident[0]. begin( ) + 1, ident[0]. end( ), input );

   element e;
   while( k )
   {
        ++ e;
        -- k;
   }
   return e;
}


void 
advance_through( token::type t, token& lookahead, ::parser::inputstream& input )
{
   if( lookahead.t != t )
   {
      throw syntaxerror( tostring( "expected ", t, 
                                   " instead of ", lookahead.t ), input );
   }
   lookahead = readtoken( input );
}


std::vector< std::string > 
readidentifier( token& lookahead, ::parser::inputstream& input )
{

   std::vector< std::string > result;

q0: 
   if( lookahead.t != token::type::ident )
   {
      throw syntaxerror( tostring( "expected identifier instead of ",
                                   lookahead.t ), input );
   }

   result. push_back( lookahead. s );
   lookahead = readtoken( input );

q1:    
   if( lookahead.t == token::type::separator )
   {
      lookahead = readtoken( input );
      goto q0;
   }

   return result;
}


set< size_t >
readassumptionset( token& lookahead, ::parser::inputstream& input )
{
   set< size_t > result; 

   advance_through( token::type::leftcurly, lookahead, input );

   if( lookahead.t != token::type::rightcurly )
   {
   q0:   
      if( lookahead.t != token::type::number )
      {   
         throw syntaxerror( tostring( "expected number, got ", 
                                      lookahead.t ), input );
      }

      if( lookahead.i >= 100000 )
      {
         throw syntaxerror( tostring( "number ", lookahead.i,
                                      " in assumption set looks a bit large" ),
                            input );
      }

      result. insert( lookahead.i );
      lookahead = readtoken( input );
     
      if( lookahead.t == token::type::comma )
      {
         lookahead = readtoken( input );
         goto q0;
      }
   }
 
   // A '}' awaits us:

   advance_through( token::type::rightcurly, lookahead, input );

   return result; 
}


truthval
readtruthval( token& lookahead, ::parser::inputstream& input )
{
   truthval res = truthval::empty; 

   advance_through( token::type::leftcurly, lookahead, input ); 

   if( lookahead.t == token::type::ident )
   {
      for( char c : lookahead.s )
      {
         truthval tv = truthval::empty; 

         switch(c)
         {
            case 'F': tv = truthval::F; break;
            case 'E': tv = truthval::E; break;
            case 'T': tv = truthval::T; break;
            default:
               throw syntaxerror( tostring( 
                                     "unrecognized truthval ", c, 
                                     " in string ",
                                     lookahead.s ), input );
         }

         // We don't accept repetititions:

         if( ( res & tv ) != truthval::empty )
            throw syntaxerror( tostring( "repeated truthval in ", 
                                         lookahead.s ), input );

         res |= tv;
      }

      lookahead = readtoken( input );
   }

   advance_through( token::type::rightcurly, lookahead, input );

   return res;
}


std::pair< std::vector< std::string >, size_t >
readpredicatedeclaration( token& lookahead, ::parser::inputstream& input )
{
   auto pred = readidentifier( lookahead, input );

   advance_through( token::type::slash, lookahead, input );

   if( lookahead.t != token::type::number )
   {
      throw syntaxerror( tostring( "expected number instead of ",
                         lookahead.t ), input ); 
   }

   size_t arity = lookahead.i;
   if( arity >= 10000 )
      throw syntaxerror( tostring( "arity of ", arity, " seems a bit large" ),
                         input );

   lookahead = readtoken( input );

   return { pred, arity }; 
}


namespace
{


   template< typename A, size_t S, 
             A R( const std::vector< std::string > & ,
                  const ::parser::inputstream& input ) >

   atom<A,S> readatom( const signature& sig,
                       token& lookahead, ::parser::inputstream& input )
   {

      if( lookahead.t == token::type::create )
      {
         // We have #-{FET}  V0.

         lookahead = readtoken( input );

         advance_through( token::type::minus, lookahead, input );

         truthval tv = readtruthval( lookahead, input );
         std::vector< std::string > arg = readidentifier( lookahead, input );
         A a = R( arg, input );

         return atom<A,S>( sig. create, tv, { a } );  
      }

      // The difficult part is to recognize V == W: 
    
      std::vector< std::string > 
              startidentifier = readidentifier( lookahead, input ); 

      if( lookahead.t == token::type::equals )
      {
         // We have V == W!

         A a1 = R( startidentifier, input );

         lookahead = readtoken( input );
         auto nextidentifier = readidentifier( lookahead, input );

         A a2 = R( nextidentifier, input );

         return atom<A,S>( sig. equality, truthval::T, { a1, a2 } ); 
      }    

      // A usual atom: 

      advance_through( token::type::minus, lookahead, input );

      truthval tv = readtruthval( lookahead, input );

      std::vector<A> arguments; 
      if( lookahead.t == token::type::leftround )
      {
         lookahead = readtoken( input );

         if( lookahead.t != token::type::rightround )
         {
         q0:
            if( lookahead.t != token::type::ident )
               throw syntaxerror( tostring( "expected identifier instead of ",
                                            lookahead.t ), input );

            auto a = readidentifier( lookahead, input );
            arguments. push_back( R( a, input ));

            if( lookahead.t == token::type::comma ) 
            {
               lookahead = readtoken( input );
               goto q0;
            }
         }

         advance_through( token::type::rightround, lookahead, input );
      }

      try
      {
         atom<A,S> res = atom<A,S>( 
                  sig. clookup( startidentifier, arguments. size( )), tv );

         for( size_t i = 0; i < arguments. size( ); ++ i )
            res[i] = arguments[i];

         return res; 
      }
      catch( const std::runtime_error& e )
      {
         // sig. clookup( ) throws an error when the name does not
         // exist. It cannot add the name, because it is const.
         // We catch it and rethrow it as syntax error. 

         std::string func = startidentifier[0];
         for( size_t i = 1; i < startidentifier. size( ); ++ i )
         {
            func += "::";
            func += startidentifier[i];
         }

         throw syntaxerror( tostring( "predicate ", 
                                      func,
                                      " has no declaration with arity ",
                                      arguments. size( )), input );
      }
   }

}


varatom readvaratom( const signature& sig,
                     token& lookahead, ::parser::inputstream& input )
{
   return readatom< variable, 6, makevariable > ( sig, lookahead, input );
}


groundatom readgroundatom( const signature& sig,
                           token& lookahead, ::parser::inputstream& input )
{
   return readatom< element, 6, makeelement > ( sig, lookahead, input );
}


std::vector< varatom >
readvaratomvector( const signature& sig,
                   token& lookahead, ::parser::inputstream& input )
{
   std::vector< varatom > vect; 

   if( lookahead.t == token::type::ident &&
       lookahead.s == "FALSE" )
   {
      lookahead = readtoken( input );
      goto accept;
   }

   if( lookahead.t != token::type::ident &&
       lookahead.t != token::type::create )
   {
      goto accept; 
   }

q0:
   vect. push_back( readvaratom( sig, lookahead, input ));
                  
   if( lookahead.t == token::type::comma )
   {
      lookahead = readtoken( input );
      goto q0;
   }

accept:
   return vect;
}


rule readrule( const signature& sig,
               token& lookahead, ::parser::inputstream& input )
{

   // The rule constructors may throw runtime errors that
   // we catch, and rethrow as conversionerrors.

   try
   {
      auto lhs = readvaratomvector( sig, lookahead, input ); 

      if( lookahead.t != token::type::bar )
      {
         // A disjunctive rule with empty rhs:

         return rule( disjunctive( lhs ), geometric::input( ));
            // geometric::input( ) is a derivation rule.
      }

      advance_through( token::type::bar, lookahead, input );

      if( lookahead.t == token::type::ident &&
          lookahead.s == "EXISTS" )
      {
         lookahead = readtoken( input );

         auto id = readidentifier( lookahead, input );
         variable var = makevariable( id, input );
            // The existentially quantified variable. 

         advance_through( token::type::colon, lookahead, input ); 

         varatom ex = readvaratom( sig, lookahead, input );

         return rule( existential( lhs, var, ex ), geometric::input( )); 
      }
      else
      {
         disjunctive disj = disjunctive( lhs ); 

         auto rhs = readvaratomvector( sig, lookahead, input ); 

         for( const auto& a : rhs )
            disj. insert_rhs(a);

         return rule( disj, geometric::input( ));
            // geometric::input( ) is a derivation rule.
      } 
   }
   catch( std::runtime_error e )
   {
      // These come out of the rule constructors.

      throw conversionerror( tostring( e. what( )), input ); 
   }

}



std::pair< interpretation, rulesystem >
                readproblem( signature& sig, ::parser::inputstream& input )
{

   // Will be rule system:

   std::vector< rule > initials;
   std::vector< rule > simplifiers;

   // Will be interpretation:

   std::vector< groundatom > atoms;
   std::vector< set< size_t >> assumptionsets;

   // Name of the problem:
 
   std::string problemname; 


   token lookahead = geometric::parser::readtoken( input );
     
   while( lookahead.t != token::type::eof )
   {
      switch( lookahead.t )
      {
      case token::type::name:
         {
            lookahead = readtoken( input );

            if( lookahead.t != token::type::quotedstring ) 
            {
               throw syntaxerror( 
                           tostring( "expected quoted string instead of ",
                                     lookahead.t ), input );
            }

            if( problemname. size( ) && problemname != lookahead.s ) 
            {
               throw syntaxerror(
                           tostring( "problem name redefined from ",
                                     problemname, " into ", 
                                     lookahead.s ), input ); 
            }

            problemname = lookahead.s;
            lookahead = readtoken( input );

            advance_through( token::type::semicolon, lookahead, input );
         }

         break;

      case token::type::signature:
         {
            lookahead = readtoken( input ); 

            while( lookahead.t == token::type::ident )
            {
               std::pair< std::vector< std::string >, size_t > 
                  decl = readpredicatedeclaration( lookahead, input );

               sig. lookup( decl. first, decl. second );  

               advance_through( token::type::semicolon, lookahead, input );
            }
         }

         break;

      case token::type::rulesystem:
      case token::type::simplifiers:
         {
            token header = lookahead;   
               // We need to know where to insert at the end. 
           
            lookahead = readtoken( input );
           
            while( lookahead.t == token::type::ident ||
                   lookahead.t == token::type::create ||
                   lookahead.t == token::type::bar ) 
            {
               rule r = readrule( sig, lookahead, input );

               advance_through( token::type::semicolon, lookahead, input ); 

               if( header.t == token::type::rulesystem )
                  initials. push_back(r);
               else
                  simplifiers. push_back(r);
            }

            break;
         }  

      case token::type::interpretation:
         {
            lookahead = readtoken( input );
       
            while( lookahead.t == token::type::ident ||
                   lookahead.t == token::type::create )
            {
               groundatom at = readgroundatom( sig, lookahead, input );

               set< size_t > a;
               if( lookahead.t == token::type::leftcurly )
               {
                  a = readassumptionset( lookahead, input );
               }

               advance_through( token::type::semicolon, lookahead, input );

               atoms. push_back( at );
               assumptionsets. push_back( a ); 
            }

            break;
         } 

      default:
         throw syntaxerror( 
            tostring( 
              "expected %name/%signature/%ruleystem/%simplifiers/"\
              "%interpretation",
              " instead of ", lookahead.t ), input );
      }
   }

   interpretation intp = interpretation{ atoms };

   for( size_t i = 0; i < atoms. size( ); ++ i )
      intp. assumptions[i] = assumptionsets[i];

   if( problemname.size( ) == 0 )
      problemname = "(no name given)";

   return { intp, { problemname, initials, simplifiers }}; 
}




}}

