
// Written by Hans de Nivelle, December 2015.

#include "tokenizer.h"
#include "../../parser/errors.h" 

namespace geometric {
namespace parser {

long unsigned int make_unsigned( std::string::const_iterator p1,
                                 std::string::const_iterator p2,
                                 const ::parser::inputstream& mess )
{
   if( p1 >= p2 )
      throw ::parser::syntaxerror( "zero-length integer", mess );

   long unsigned int val = 0;
   while( p1 < p2 )
   {
      char c = *p1;
      if( ! isdigit(c) )
         throw ::parser::syntaxerror( "non-digit in integer", mess );

      char x = c - '0';

      val = val * 10;
      val += x;

      ++ p1;
   }
   return val;
}


token readtoken( ::parser::inputstream& stream )
{

whitespace:
   if( stream. eof( ))
      return token( token::type::eof );

   char c = stream. getnextchar( ); 

   if( iswhitespace(c))
   {
      stream. moveforward( );
      goto whitespace; 
   }

   if( isdigit(c))
   {
      std::string s = {c};
      stream. moveforward( );

      while( !stream. eof( ) && isdigit( stream. getnextchar( )))
      {
         s. push_back( stream. getnextchar( ));
         stream. moveforward( );
      }

      if( s. size( ) > 15 )
      {
         throw ::parser::syntaxerror( 
                               tostring( "number is too big: ", s ), stream );
      }

      return token( token::type::number, 
                    make_unsigned( s. begin( ), s. end( ), stream ));
   }

   if( startidentifier(c))
   {
      std::string s = {c};
      stream. moveforward( );

      while( !stream. eof( ) && inidentifier( stream. getnextchar( )))
      {
         s. push_back( stream. getnextchar( ));
         stream. moveforward( );
      }

      return token( token::type::ident, s );
   }    

   if( c == '|' )
   {
      stream. moveforward( );
      return token( token::type::bar );
   }

   if( c == ':' )
   {
      stream. moveforward( );
      if( !stream. eof( ) && stream. getnextchar( ) == ':' )
      {
         stream. moveforward( );
         return token( token::type::separator );
      }
      else
         return token( token::type::colon );
   }

   if( c == '(' )
   {
      stream. moveforward( );
      return token( token::type::leftround );
   }

   if( c == ')' )
   {
      stream. moveforward( );
      return token( token::type::rightround );
   }

   if( c == '{' )
   {
      stream. moveforward( );
      return token( token::type::leftcurly );
   }

   if( c == '}' )
   {
      stream. moveforward( );
      return token( token::type::rightcurly );
   }

   if( c == '[' )
   {
      stream. moveforward( );
      return token( token::type::leftsquare );
   }

   if( c == ']' )
   {
      stream. moveforward( );
      return token( token::type::rightsquare );
   }

   if( c == '=' )
   {
      stream. moveforward( );
      if( !stream. eof( ) && stream. getnextchar( ) == '=' )
      {
         stream. moveforward( ); 
         return token( token::type::equals );
      }

      throw ::parser::syntaxerror(
         std::string( "unrecognized character '='" ), stream );
   }

   if( c == '#' )
   {
      stream. moveforward( );
      return token( token::type::create );
   }

   if( c == ',' )
   {
      stream. moveforward( );
      return token( token::type::comma );
   }

   if( c == ';' )
   {
      stream. moveforward( );
      return token( token::type::semicolon );
   }

   if( c == '+' )
   {
      stream. moveforward( );
      return token( token::type::plus );
   }

   if( c == '-' )
   {
      stream. moveforward( );
      return token( token::type::minus );
   }

   if( c == '/' )
   {
      // Either: 
      //    a single slash:             /
      //    a comment:                  /*                 */
      //    a comment:                  //

      stream. moveforward( );

      if( !stream. eof( ) && stream. getnextchar( ) == '/' )
      {
         stream. moveforward( );
         while( !stream. eof( ) && stream. getnextchar( ) != '\n' )
            stream. moveforward( );

         goto whitespace; 
      }

      if( !stream. eof( ) && stream. getnextchar( ) == '*' )
      {
         stream. moveforward( );

         int q = 0;                // State of DFA that recognizes '*/'.
         while( q != 2 )
         {
            if( stream. eof( ))
               throw ::parser::syntaxerror( 
                  "end-of-file encountered in comment", stream );

            char c = stream. getnextchar( );
            if( q == 0 && c == '*' )
               q = 1;
            else
            {
               if( q == 1 && c == '/' )
                  q = 2;
               else
                  q = 0;
            }
            stream. moveforward( );
         }
         goto whitespace; 
      }

      return token( token::type::slash );
   }

   if( c == '.' )
   {
      stream. moveforward( );
      return token( token::type::dot );
   }

   if( c == '\'' )
   {
      stream. moveforward( );

      std::string s;

      while( !stream. eof( ) && 
                 stream. getnextchar( ) != '\'' &&
                 stream. getnextchar( ) != '\n' ) 
      {
         s. push_back( stream. getnextchar( ));
         stream. moveforward( );  
      }

      if( stream. eof( ))
      {
         throw ::parser::syntaxerror(
            "end-of-file encountered in quoted string", stream );
      }

      if( stream. getnextchar( ) == '\n' )
      {
         throw ::parser::syntaxerror(
            "newline encountered in quoted string", stream );
      }

      // It must be a '

      stream. moveforward( );
      return token( token::type::quotedstring, s );
   }
      
   if( c == '%' )
   {
      stream. moveforward( );
      while( !stream. eof( ) && iswhitespace( stream. getnextchar( )))
         stream. moveforward( );

      std::string s;
      while( !stream. eof( ) && startidentifier( stream. getnextchar( ))) 
      {
         s. push_back( stream. getnextchar( ));
         stream. moveforward( );
      }

      if( s == "name" )
         return token( token::type::name );
      if( s == "signature" )
         return token( token::type::signature );
      if( s == "rulesystem" )
         return token( token::type::rulesystem );
      if( s == "simplifiers" )
         return token( token::type::simplifiers );
      if( s == "interpretation" )
         return token( token::type::interpretation );

      if( s == "eof" )
         return token( token::type::eof );

      if( s. size( ) == 0 )
         throw ::parser::syntaxerror( "no header found after '%'", stream );

      throw ::parser::syntaxerror( 
                tostring( "unrecognized header ", s, " after '%'" ), stream );   
   }
 
   throw ::parser::syntaxerror( 
                tostring( "unrecognized character '", c, "'" ), stream );    
}


void testtokenizer( )
{
   try
   {
      ::parser::inputstream is{ std::cin, "STDIN" };

      auto t = geometric::parser::readtoken( is );
      std::cout << t << "\n";
      while( t.t != geometric::parser::token::type::eof )
      {
         t = geometric::parser::readtoken( is );
         std::cout << t << "\n";
      }
   }
   catch( ::parser::syntaxerror synt )
   {
      std::cout << synt << "\n";
   }

}


}}


